from naas_drivers.driver import InDriver, OutDriver

from typing import Any, Dict, List
import pandas as pd
import requests
import datetime
import shutil
import json
import time
import uuid
import os
import io
from IPython.core.display import display, HTML
import cson
import jwt


class Toucan(InDriver, OutDriver):
    """Toucan lib"""

    # Private vars
    __url_screenshot_api = os.environ.get(
        "TC_API_SCREENSHOT", "http://toucan-screenshot:3000/screenshot"  # Sensitive
    )
    __TOUCAN_EMBED_ENCRYPTION_KEY = os.environ.get("TOUCAN_EMBED_ENCRYPTION_KEY", None)
    __url_config = "config"
    __url_load = "load"
    __url_embed = "embed"
    __url_release = "release"
    __url_reports = "reports"
    __url_data = "data"
    __url_login = "login"
    __ur_ids = "ids"
    __url_users = "users"
    __url_small_apps = "small-apps"
    __url_tc_params = "scripts/tc-params.js"
    __url_tc_app_version = "tucana-version.txt"
    __replace_tc_params = "window.TC_PARAMS = "
    __token = None
    __url_name = None
    # Public vars
    url_base = None
    url_api = None
    login = {}
    user = {}
    config = {}
    tc_params = {}
    small_apps = []
    debug = False

    def connect(self, url, username, password, debug=False):
        self.url_base = url
        host = url.partition("://")[2]
        self.__url_name = host.partition(".")[0]
        self.login = {"username": username, "password": password}
        self.debug = debug
        try:
            if self.debug:
                print("Get toucan tc-params")
            self.tc_params = self.__request_tc_params()
            if self.debug:
                print("Get toucan url_api")
            self.url_api = self.tc_params.get("API_BASEROUTE")
            if self.debug:
                print("Get toucan user")
            self.user = self.__request_user()
            if self.debug:
                print("Get toucan token")
            self.__token = self.user.get("token")
            if self.debug:
                print("Get toucan small_apps")
            self.small_apps = self.__request_small_apps()
        except requests.exceptions.HTTPError as errh:
            print("Http Error:", errh.response.text)
        except requests.exceptions.ConnectionError as errc:
            print("Error Connecting:", errc)
        except requests.exceptions.Timeout as errt:
            print("Timeout Error:", errt)
        except requests.exceptions.RequestException as err:
            print("OOps: Something Else", err)
        if self.debug:
            print("user logged")
        self.connected = True
        return self

    def __request_small_apps(self):
        req = requests.get(
            f"{self.url_api}/{self.__url_small_apps}", headers=self.__get_headers()
        )
        req.raise_for_status()
        return req.json()

    def __request_tc_params(self):
        req = requests.get(f"{self.url_base}/{self.__url_tc_params}")
        req.raise_for_status()
        res = (
            req.text.replace(self.__replace_tc_params, "")
            .replace("\t", " ")
            .replace("\n", "")
            .replace(";", "")
        )
        json_conf = json.loads(res)
        return json_conf

    def __request_user(self):
        req = requests.post(f"{self.url_api}/{self.__url_login}", json=self.login)
        req.raise_for_status()
        return req.json()

    def __createFolder(self, name):
        try:
            os.makedirs(name)
            print(f"Successfully created the directory {name}")
        except OSError:
            print(f"Directory {name} already exist")

    def __calc_report_id(self, current_id, ids):
        calc_id = current_id
        if ids is not None and 0 in ids:
            calc_id += int(ids.get(0))
        return calc_id

    def __filterReport(self, reports, report_name):
        res_report = [{"id": 0, "entityName": "default"}]
        if report_name is None:
            return reports
        if not reports or len(reports) == 0:
            return res_report
        found = self.get_report_by_name(reports, report_name)
        if found:
            res_report = [found]
        return res_report

    def __generateUrls(self, app_name, config, reports, ids):
        arr = []
        if self.debug:
            print("Get app slides")
        slides = config.get("slides")
        if self.debug:
            print("Get app Home")
        home = config.get("home")
        if home is not None and "skipToReport" in home:
            arr.append(
                {
                    "url": f"{self.url_base}/{app_name}",
                    "selector": self.get_dashboard_selector(),
                    "name": "dashboard",
                }
            )
            if self.debug:
                print("Add New dashboard to url list")
        else:
            arr.append(
                {
                    "url": f"{self.url_base}/{app_name}",
                    "selector": ".report-execsum",
                    "name": "dashboard",
                }
            )
            if self.debug:
                print("Add Old dashboard to url list")
        if self.debug:
            print("Generate all url from config")
        for report in reports:
            if "id" in report and "entityName" in report:
                for slide in slides:
                    if "parent_id" in slide and "id" in slide and "title" in slide:
                        name = self.generate_screenshot_name(slide, report)
                        r_id = self.__calc_report_id(report.get("id"), ids)
                        if self.debug:
                            print("Generate url for", name)
                        arr.append(
                            {
                                "url": f"{self.url_base}/{app_name}?report={r_id}&dashboard={r_id}&slide={slide.get('id')}",
                                "selector": ".tc-slide__content, .tc-story",
                                "name": name,
                            }
                        )
        return arr

    def __get_headers(self):
        return {"authorization": f"Bearer {self.__token}"}

    def embed(self, small_app, slide, hosts=None, mode="webcomponent", height="800px"):
        allowed_hosts = (
            hosts
            if hosts
            else [
                os.environ.get("PUBLIC_PROXY_API", ""),
                os.environ.get("JUPYTERHUB_URL", ""),
            ]
        )
        uid = str(uuid.uuid4())
        data = {
            "allowedHosts": allowed_hosts,
            "expirationDate": None,
            "layout": {
                "type": "single",
                "content": {"path": f"slides[?id==`{slide}`]", "variables": {}},
            },
            "paths": [f"slides[?id==`{slide}`]"],
            "public": False,
            "smallApp": small_app,
            "uid": uid,
        }
        req = requests.post(
            f"{self.url_api}/{self.__url_embed}",
            headers=self.__get_headers(),
            json=data,
        )
        req.raise_for_status()
        html = None
        if mode == "webcomponent":
            html = f"""
                <div style="height:{height}">
                <script async src="{self.url_base}/scripts/embedLauncher.js?id={uid}&token={self.__token}" type="text/javascript"></script>
                </div>
            """
        else:
            html = f"""
                <div style="height:{height}">
                <iframe style="border: 0; overflow: hidden;" frameBorder="0" height="100%" width="100%"
                    src="{self.url_base}/embed.html?id=id={uid}&token={self.__token}"></iframe>
                </div>
            """
        display(HTML(html))
        return html

    def craft_token(
        self,
        username: str,
        small_apps_access: Dict[str, str],
        groups: List[str] = None,
        extra_infos: Dict[str, Any] = None,
        expires_in: datetime.timedelta = datetime.timedelta(hours=1),
    ) -> str:
        user_payload = {
            "username": username,
            "roles": ["USER"],
            "privileges": {"smallApp": small_apps_access},
            "groups": groups,
            "attributes": extra_infos,
        }
        payload = {
            **user_payload,
            "iat": datetime.datetime.utcnow(),
            "exp": datetime.datetime.utcnow() + expires_in,
        }
        return jwt.encode(
            payload, self.__TOUCAN_EMBED_ENCRYPTION_KEY, algorithm="HS256"
        ).decode("utf8")

    def get_dashboard_selector(self):
        status_app = self.get_version()
        version = status_app["frontVersion"]
        if version == "v64.0.0":
            return ".dashboard-report"
        else:
            return ".small-app-home__content"

    def get_version(self):
        req_api = requests.get(f"{self.url_api}", headers=self.__get_headers())
        req_api.raise_for_status()
        result = req_api.json()
        req_app = requests.get(f"{self.url_base}/{self.__url_tc_app_version}")
        req_app.raise_for_status()
        front_version = req_app.text().strip()
        result["frontVersion"] = front_version
        return result

    def get_app_config(self, app_name):
        req = requests.get(
            f"{self.url_api}/{app_name}/{self.__url_config}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def get_app_data(self, app_name):
        req = requests.get(
            f"{self.url_api}/{app_name}/{self.__url_data}", headers=self.__get_headers()
        )
        req.raise_for_status()
        return req.json()

    def get_app_reports_ids(self, app_name):
        req = requests.get(
            f"{self.url_api}/{app_name}/{self.__url_reports}/{self.__ur_ids}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def get_app_reports(self, app_name):
        req = requests.get(
            f"{self.url_api}/{app_name}/{self.__url_reports}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def get_users(self):
        req = requests.get(
            f"{self.url_api}/{self.__url_users}", headers=self.__get_headers()
        )
        req.raise_for_status()
        return req.json()

    def get_users_analytics(self):
        users = self.get_users()
        flatten_users = []
        for user in users:
            if user.get("logins"):
                for login in user.get("logins"):
                    date = time.ctime(login.get("date").get("$date") / 1000)
                    TIME = datetime.datetime.strptime(
                        date, "%a %b %d %H:%M:%S %Y"
                    ).strftime("%H:%M:%S")
                    DATE = datetime.datetime.strptime(
                        date, "%a %b %d %H:%M:%S %Y"
                    ).strftime("%Y-%m-%d")
                    flatten = {"USER": user.get("username"), "TIME": TIME, "DATE": DATE}
                    flatten_users.append(flatten)
            else:
                flatten = {"USER": user.get("username"), "TIME": None, "DATE": None}
                flatten_users.append(flatten)

        df = (
            pd.DataFrame(flatten_users)
            .drop(["TIME"], axis=1)
            .drop_duplicates()
            .reset_index(drop=True)
        )
        df["VALUE"] = 1

        df_session = df.copy()
        df_session = df_session.groupby(["USER", "DATE"], as_index=False).count()

        df_session_all = df.copy()
        df_session_all = (
            df_session_all.drop(["USER"], axis=1)
            .groupby(["DATE"], as_index=False)
            .count()
        )
        df_session_all["USER"] = "All users"

        df_final = pd.concat([df_session, df_session_all], axis=0)
        return df_final

    def send_app_config(
        self, app_name, file_upload, format_file="front_config", stage="staging"
    ):
        config_name = ""
        format_url = "&format=cson"
        files = None
        data = None
        if format_file == "front_config":
            files = {"file": io.StringIO(json.dumps({"json": cson.load(file_upload)}))}
        elif format_file == "etl_config":
            config_name = "/etl"
            data = file_upload
        elif format_file == "report":
            config_name = "/report"
            data = file_upload
        elif format_file == "dashboard-Group":
            config_name = "/dashboard-Group"
            data = file_upload
        elif format_file == "augment.py":
            config_name = "/augment"
            files = {"file": file_upload}
            format_url = ""
        elif format_file == "preprocess_validation":
            config_name = "/preprocess_validation"
            data = file_upload
        elif format_file == "permissions.py":
            config_name = "/permissions"
            files = {"file": file_upload}
            format_url = ""
        elif format_file == "permissions_config":
            config_name = "/permissions_config"
            data = file_upload
        elif format_file == "notifications_handlers.py":
            config_name = "/notifications_handlers"
            files = {"file": file_upload}
            format_url = ""

        req = requests.put(
            f"{self.url_api}/{app_name}/{self.__url_config}{config_name}?stage={stage}{format_url}",
            headers=self.__get_headers(),
            files=files,
            data=data,
        )
        req.raise_for_status()
        return req.json()

    def download_app_config(
        self, app_name, format_file="front_config", stage="staging"
    ):
        format_url = ""
        config_name = ""
        format_file = "&format=cson"
        if format_file == "etl_config":
            config_name = "/etl"
        elif format_file == "report":
            config_name = "/report"
        elif format_file == "dashboard-Group":
            config_name = "/dashboard-Group"
        elif format_file == "augment.py":
            config_name = "/augment"
            format_url = ""
        elif format_file == "preprocess_validation":
            config_name = "/preprocess_validation"
        elif format_file == "permissions.py":
            config_name = "/permissions"
            format_url = ""
        elif format_file == "permissions_config":
            config_name = "/permissions_config"
        elif format_file == "notifications_handlers.py":
            config_name = "/notifications_handlers"
            format_url = ""

        req = requests.get(
            f"{self.url_api}/{app_name}/{self.__url_config}{config_name}?stage={stage}{format_url}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req

    def deploy_app(
        self,
        app_name,
        stage="staging",
        operations=[
            "preprocess_data_sources",
            "populate_reports",
            "populate_dashboards",
            "release_design",
        ],
        force=True,
    ):
        force_str = "true" if force else "false"
        req = requests.post(
            f"{self.url_api}/{app_name}/{self.__url_config}/pull?force={force_str}&stage={stage}",
            headers=self.__get_headers(),
            json={"operations": operations},
        )
        req.raise_for_status()
        return req.json()

    def get_data(self, app_name, domain, stage="staging"):
        req = requests.get(
            f"{self.url_api}/{app_name}/domain/{domain}?stage={stage}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        s_data = str(req.content, "utf-8")
        return io.StringIO(s_data)

    def get_metadata(self, app_name, stage="staging"):
        req = requests.get(
            f"{self.url_api}/{app_name}/metadata?stage={stage}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def create_small_app(self, app_name, id_app=None):
        body = {"name": app_name, "id": id_app if id_app else app_name}
        req = requests.post(
            f"{self.url_api}/small-apps", headers=self.__get_headers(), json=body
        )
        req.raise_for_status()
        return req.json()

    def load_operations(
        self,
        app_name,
        stage="staging",
        notification=False,
        operations=[
            "preprocess_data_sources",
            "populate_basemaps",
            "populate_reports",
            "populate_dashboards",
            "populate_permissions",
        ],
    ):
        notification_str = "true" if notification else "false"
        req = requests.post(
            f"{self.url_api}/{app_name}/{self.__url_config}/operations?notify={notification_str}&stage={stage}",
            headers=self.__get_headers(),
            json={"operations": operations},
        )
        req.raise_for_status()
        return req.json()

    def load_conf(self, app_name, stage="staging"):
        req = requests.post(
            f"{self.url_api}/{app_name}/{self.__url_load}?stage={stage}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def release_conf(self, app_name, stage="staging"):
        req = requests.post(
            f"{self.url_api}/{app_name}/{self.__url_release}?stage={stage}",
            headers=self.__get_headers(),
        )
        req.raise_for_status()
        return req.json()

    def get_report_by_name(self, reports, report_name):
        report_found = None
        if not reports or len(reports) == 0:
            return report_found
        for report in reports:
            if (
                report_name is not None
                and "id" in report
                and "entityName" in report
                and report.get("entityName") == report_name
            ):
                report_found = report
        return report_found

    def screenshots_app_all(self):
        for app in self.small_apps:
            self.screenshots_app(app)

    def generate_screenshot_name(self, slide, report):
        title = slide.get("title", "noTitle")
        title = "".join(x for x in title if x.isalnum())
        name = report.get("entityName", "noEntity")
        name = "".join(x for x in name if x.isalnum())
        return f"{name}_{title}"

    def generate_folder_name(self, app_name):
        today = datetime.date.today().strftime("%Y_%m_%d")
        return f"{today}/{self.__url_name}/{app_name}"

    def is_app_allowed(self, app_name):
        res = False
        for small_app in self.small_apps:
            if small_app.get("id") == app_name:
                res = True
        return res

    def screenshots_app(self, app_name, report_name=None, full_size=False):
        if self.debug:
            print("Generate screenshots for", app_name)
        if self.is_app_allowed(app_name):
            folder_name = self.__generate_folder_name(app_name)
            self.__createFolder(folder_name)
            if self.debug:
                print("Get app config")
            config = self.get_app_config(app_name)
            ids = None
            try:
                ids = self.get_app_reports_ids(app_name)
            except requests.exceptions.RequestException:
                print("Cannot get report ids")
            reports_all = self.get_app_reports(app_name)
            reports = self.__filterReport(reports_all, report_name)
            urls = self.__generateUrls(app_name, config, reports, ids)
            for url in urls:
                req = ""
                path = f"{folder_name}/{url.get('name')}.png"
                try:
                    if self.debug:
                        print(f"Request Screenshot {path}")
                    req = requests.get(
                        self.__url_screenshot_api,
                        params={
                            "url": url.get("url"),
                            "format": "png",
                            "lang": "en",
                            "fullSize": full_size,
                            "deviceScaleFactor": 1,
                            "token": self.__token,
                            "elementSelector": url.get("selector"),
                        },
                        stream=True,
                    )
                    req.raise_for_status()
                except requests.exceptions.HTTPError as errh:
                    print("Http Error:", errh.response.text)
                except requests.exceptions.ConnectionError as errc:
                    print("Error Connecting:", errc)
                except requests.exceptions.Timeout as errt:
                    print("Timeout Error:", errt)
                except requests.exceptions.RequestException as err:
                    print("OOps: Something Else", err)
                path = f"{folder_name}/{url.get('name')}.png"
                if req.status_code == 200:
                    with open(path, "wb") as out_file:
                        shutil.copyfileobj(req.raw, out_file)
                        if self.debug:
                            print(f"Created Screenshot {path}")
        else:
            print("this app is not allow to your account")



from naas_drivers.driver import InDriver, OutDriver
import dataclasses
from dataclasses import dataclass, field
from typing import List, Optional, Union
from numbers import Number
from enum import Enum
import json
import logging
from copy import deepcopy
import pandas as pd
import threading
import pydash
from notion_client import Client
from dacite import from_dict


VERSION = "2021-08-16"

global notion_instance


def ensure_page_id(page_id):
    if "http" in page_id:
        page_id = page_id.split("/")[-1].split("-")[-1].split("?")[0]
    return page_id


def ensure_database_id(database_id):
    if "http" in database_id:
        database_id = database_id.split("/")[-1].split("?")[0]
    return database_id


class Notion(InDriver, OutDriver):
    __client = None

    def __init__(self):
        """
        Classes / Methods to query notion
        """
        self.database = self.Databases(self)
        self.databases = self.database

        self.page = self.Pages(self)
        self.pages = self.page

        self.block = self.Blocks(self)
        self.blocks = self.block

        self.user = self.Users(self)
        self.users = self.user

        """
        Data classes representing Page/Database/Blocks.
        """
        self.Page = Page
        self.Database = Database
        self.Block = Block

        """
        Global to allow all dataclasses to access notion client and make requests
        from a dataclass directly.
        """
        global notion_instance
        notion_instance = self

    @staticmethod
    def instance():
        return notion_instance

    @property
    def client(self):
        return self.__client

    def connect(self, token: str, log_level: int = logging.WARNING):
        self.__client = Client(auth=token, log_level=log_level)
        return self

    def to_json(self, source):
        return json.dumps(source, cls=EnhancedJSONEncoder, ensure_ascii=False)

    def to_dict(self, source):
        return del_none(json.loads(self.to_json(source)))

    class __InnerBase:
        __parent = None

        def __init__(self, parent):
            self.__parent = parent

        @property
        def client(self):
            return self.__parent.client

        @property
        def parent(self):
            return self.__parent

    class Databases(__InnerBase):
        def __from_dict(self, data):
            return from_dict(data_class=Database, data=data)

        def query(self, database_id, query={}):
            database_id = ensure_database_id(database_id)
            ret = []
            has_more = True
            next_cursor = None

            while has_more:
                if has_more is True and next_cursor is not None:
                    query["start_cursor"] = next_cursor

                response = self.client.databases.query(database_id=database_id, **query)
                results = response.get("results")

                for r in results:
                    ret.append(from_dict(data_class=Page, data=r))

                if response.get("has_more") is True:
                    has_more = True
                    next_cursor = response.get("next_cursor")
                else:
                    has_more = False

            return ret

        def create(self, db):
            payload = self.parent.to_dict(db)
            return self.__from_dict(self.client.databases.create(**payload))

        def update(self, db):
            payload = self.parent.to_dict(db)
            return self.__from_dict(
                self.client.databases.update(database_id=db.id, **payload)
            )

        def get(self, database_id):
            return self.retrieve(database_id)

        def retrieve(self, database_id):
            database_id = ensure_database_id(database_id)
            raw = self.client.databases.retrieve(database_id)
            return self.__from_dict(raw)

    class Pages(__InnerBase):
        def __from_dict(self, data):
            return from_dict(data_class=Page, data=data)

        def create_from_page(self, page):
            payload = self.parent.to_dict(page)
            return self.__from_dict(self.client.pages.create(**payload))

        def create(self, database_id, title):
            database_id = ensure_database_id(database_id)
            new_page = Page.new(database_id=database_id)
            new_page.title("Name", title)
            return new_page.create()

        def retrieve(self, page_id: str):
            page_id = ensure_page_id(page_id)
            page = self.client.pages.retrieve(page_id=page_id)
            return self.__from_dict(page)

        def get(self, page_id: str):
            return self.retrieve(page_id)

        def update(self, page):  # TODO: Improve readability.
            copied = deepcopy(page)
            filtered_properties = {}
            for p in copied.properties:
                prop = copied.properties[p]
                if (
                    prop.type
                    not in [
                        "last_edited_by",
                        "last_edited_time",
                        "created_by",
                        "created_time",
                    ]
                    and getattr(prop, prop.type) is not None
                    and not (
                        prop.type == "rich_text"
                        and pydash.get(getattr(prop, prop.type), "[0].plain_text") == ""
                    )
                ):
                    filtered_properties[p] = prop
            copied.properties = filtered_properties
            payload = self.parent.to_dict(copied)
            return self.__from_dict(
                self.client.pages.update(page_id=copied.id, **payload)
            )

    class Blocks(__InnerBase):
        def __from_dict(self, data):
            return from_dict(data_class=Block, data=data)

        def retrieve(self, block_id: str):
            block = self.client.blocks.retrieve(block_id=block_id)
            return self.__from_dict(block)

        def get(self, block_id: str):
            return self.retrieve(block_id)

        def children(self, block_id: str):
            ret = []
            blocks = self.client.blocks.children.list(block_id=block_id).get("results")
            for r in blocks:
                ret.append(self.__from_dict(r))
            return ret

        def append(self, block_id, blocks):
            if type(blocks) != list:
                blocks = [blocks]
            payload = self.parent.to_dict(blocks)
            ret = []
            for b in self.client.blocks.children.append(
                block_id=block_id, children=payload
            ).get("results"):
                ret.append(self.__from_dict(b))
            return ret

        def update(self, block):
            payload = self.parent.to_dict(block)
            return self.__from_dict(
                self.client.blocks.update(block_id=block.id, **payload)
            )

        def delete(self, block_id):
            return self.__from_dict(self.client.blocks.delete(block_id=f"{block_id}"))

    class Users(__InnerBase):
        def __from_dict(self, data):
            return from_dict(data_class=User, data=data)

        def retrieve(self, user_id):
            return self.__from_dict(self.client.users.retrieve(user_id=user_id))

        def get(self, user_id):
            return self.retrieve(user_id)

        def list(self):
            ret = []
            for b in self.client.users.list().get("results"):
                ret.append(self.__from_dict(b))
            return ret


"""Notion Data classes"""


class __BaseDataClass:
    """
    __BaseDataClass is used to make it easy from any @dataclass to access notion client.
    It allows to update/create/delete/query objects directly from a dataclass.
    """

    def to_json(self, source=None):
        if source is None:
            source = self
        return json.dumps(source, cls=EnhancedJSONEncoder)

    @property
    def notion(self):
        return Notion.instance()


@dataclass
class FileExternal:
    url: Optional[str] = ""

    @classmethod
    def new(cls, url: str) -> "FileExternal":
        return cls(url=url)


@dataclass
class File(__BaseDataClass):
    url: Optional[str]
    expiry_time: Optional[str]
    name: Optional[str]
    external: Optional[FileExternal] = field(default_factory=FileExternal)
    type: str = "external"

    @classmethod
    def new(cls, url: str) -> "File":
        return cls(
            url=None, expiry_time=None, name=None, external=FileExternal(url=url)
        )


@dataclass
class Emoji(__BaseDataClass):
    emoji: str
    type: str = "emoji"

    @classmethod
    def new(cls, emoji: str) -> "Emoji":
        return cls(emoji=emoji)


@dataclass
class Parent(__BaseDataClass):
    type: str
    workspace: Optional[bool] = None
    page_id: Optional[str] = None
    database_id: Optional[str] = None

    @classmethod
    def new(cls, type: str, value: Union[bool, str]) -> "Parent":
        if type == "page_id":
            return cls(type=type, page_id=value)
        elif type == "database_id":
            return cls(type=type, database_id=value)
        elif type == "workspace":
            return cls(type=type, workspace=value)
        else:
            return cls(type=type)

    @classmethod
    def new_page_parent(cls, page_id):
        return cls(type="page_id", page_id=page_id)

    @classmethod
    def new_workspace_parent(cls):
        return cls(type="workspace", workspace=True)

    @classmethod
    def new_database_parent(cls, database_id):
        return cls(type="database_id", database_id=database_id)


@dataclass
class Link(__BaseDataClass):
    url: str
    type: str = "url"

    @classmethod
    def new(cls, url: str) -> "Link":
        return cls(url=url)


@dataclass
class Text(__BaseDataClass):
    content: str
    link: Optional[Link] = None

    @classmethod
    def new(cls, content: str) -> "Text":
        return cls(content=content)


@dataclass
class Annotation(__BaseDataClass):
    bold: bool = False
    italic: bool = False
    strikethrough: bool = False
    underline: bool = False
    code: bool = False
    color: str = "default"

    @classmethod
    def new(cls) -> "Annotation":
        return cls()


@dataclass
class Person(__BaseDataClass):
    email: str

    @classmethod
    def new(cls, email: str) -> "Person":
        return cls(email=email)


@dataclass
class User(__BaseDataClass):
    object: str
    id: str
    type: Optional[str]
    name: Optional[str]
    avatar_url: Optional[str]
    person: Optional[Person]
    bot: Optional[object]

    @classmethod
    def new(cls, user_id: str) -> "User":
        return notion_instance.users.get(user_id)

    def __post_init__(self):
        if self.type == "person":
            self.person = Person("")
        if self.type == "bot":
            self.bot = {}


@dataclass
class Date(__BaseDataClass):
    start: str
    end: Optional[str]

    @classmethod
    def new(cls, start: str) -> "Date":
        return cls(start=start)


@dataclass
class Equation(__BaseDataClass):
    expression: str

    @classmethod
    def new(cls, expression: str) -> "Equation":
        return from_dict(data_class=cls, expression=expression)


@dataclass
class UserMention(__BaseDataClass):
    type: str
    user: User


@dataclass
class RichText(__BaseDataClass):
    plain_text: str
    href: Optional[str] = None
    annotations: Annotation = field(default_factory=Annotation)
    type: str = "text"
    text: Optional[Text] = None
    mention: Optional[UserMention] = None
    page: Optional[object] = None
    database: Optional[object] = None
    date: Optional[Date] = None
    equation: Optional[Equation] = None

    @classmethod
    def new(cls, text: str) -> "RichText":
        # Notion does not accept empty text anymore, we added spaces to fix for now (2022-02-23)
        if text == "":
            return cls(type="text", plain_text=" ", text=Text.new(" "))
        return cls(type="text", plain_text=text, text=Text.new(text))

    @classmethod
    def new_text(cls, content):
        # Notion does not accept empty text anymore, we added spaces to fix for now (2022-02-23)
        if content == "":
            return cls(plain_text=" ", text=Text(content=" "))
        return cls(plain_text=content, text=Text(content=content))


"""
Database and Database properties

This section is containing all classes related to Properties (https://developers.notion.com/reference/database#property-object)
"""


@dataclass
class Property(__BaseDataClass):
    id: Optional[str] = None
    type: str = None  # TODO: Consider using enum to improve typing.


@dataclass
class DatabaseProperty(Property):
    name: str = None


@dataclass
class DatabasePropertyTitle(DatabaseProperty):
    title: object = field(default_factory=dict)
    type: str = "title"


@dataclass
class DatabasePropertyText(DatabaseProperty):
    rich_text: object = field(default_factory=dict)
    type: str = "rich_text"


@dataclass
class DatabasePropertyNumber_Configuration:
    format: str = "number"


@dataclass
class DatabasePropertyNumber(DatabaseProperty):
    number: DatabasePropertyNumber_Configuration = field(default_factory=dict)
    type: str = "number"


@dataclass
class DatabasePropertySelect_Configuration_Option:
    name: str
    id: str
    color: str  # TODO: enum


@dataclass
class DatabasePropertySelect_Configuration:
    options: List[DatabasePropertySelect_Configuration_Option] = field(
        default_factory=list
    )


@dataclass
class DatabasePropertySelect(DatabaseProperty):
    select: DatabasePropertySelect_Configuration = field(
        default_factory=DatabasePropertySelect_Configuration
    )
    type: str = "select"

    def get_option_by_name(self, name):
        if self.select and self.select.options:
            for option in self.select.options:
                if option.name == name:
                    return option
        return None


@dataclass
class DatabasePropertyMultiSelect(DatabaseProperty):
    multi_select: DatabasePropertySelect_Configuration = field(
        default_factory=DatabasePropertySelect_Configuration
    )  # TODO: Create own childs? Not reuse DatabasePropertySelect_Configuration class.
    type: str = "multi_select"

    def get_option_by_name(self, name):
        if self.multi_select and self.multi_select.options:
            for option in self.multi_select.options:
                if option.name == name:
                    return option
        return None


@dataclass
class DatabasePropertyDate(DatabaseProperty):
    date: object = field(default_factory=dict)
    type: str = "date"


@dataclass
class DatabasePropertyPeople(DatabaseProperty):
    people: object = field(default_factory=dict)
    type: str = "people"


@dataclass
class DatabasePropertyFiles(DatabaseProperty):
    files: object = field(default_factory=dict)
    type: str = "files"


@dataclass
class DatabasePropertyCheckbox(DatabaseProperty):
    checkbox: object = field(default_factory=dict)
    type: str = "checkbox"


@dataclass
class DatabasePropertyUrl(DatabaseProperty):
    url: object = field(default_factory=dict)
    type: str = "url"


@dataclass
class DatabasePropertyEmail(DatabaseProperty):
    email: object = field(default_factory=dict)
    type: str = "email"


@dataclass
class DatabasePropertyPhoneNumber(DatabaseProperty):
    phone_number: object = field(default_factory=dict)
    type: str = "phone_number"


@dataclass
class DatabasePropertyFormula_Configuration:
    expression: str = ""
    type: str = "expression"


@dataclass
class DatabasePropertyFormula(DatabaseProperty):
    formula: DatabasePropertyFormula_Configuration = field(
        default_factory=DatabasePropertyFormula_Configuration
    )
    type: str = "formula"


@dataclass
class DatabasePropertyRelation_Configuration:
    database_id: str
    synced_property_name: Optional[str]
    synced_property_id: Optional[str]


@dataclass
class DatabasePropertyRelation(DatabaseProperty):
    relation: DatabasePropertyRelation_Configuration = field(
        default_factory=DatabasePropertyRelation_Configuration
    )
    type: str = "relation"


@dataclass
class DatabasePropertyRollup_Configuration:
    relation_property_name: str
    relation_property_id: str
    rollup_property_name: str
    rollup_property_id: str
    function: str


@dataclass
class DatabasePropertyRollup(DatabaseProperty):
    rollup: DatabasePropertyRollup_Configuration = field(
        default_factory=DatabasePropertyRollup_Configuration
    )
    type: str = "rollup"


@dataclass
class DatabasePropertyCreatedBy(DatabaseProperty):
    created_by: object = field(default_factory=dict)
    type: str = "created_by"


@dataclass
class DatabasePropertyCreatedTime(DatabaseProperty):
    created_time: object = field(default_factory=dict)
    type: str = "created_time"


@dataclass
class DatabasePropertyLastEditedTime(DatabaseProperty):
    last_edited_time: object = field(default_factory=dict)
    type: str = "last_edited_time"


@dataclass
class DatabasePropertyLastEditedBy(DatabaseProperty):
    last_edited_by: object = field(default_factory=dict)
    type: str = "last_edited_by"


class DatabasePropertyFactory:
    """
    This class is a helper to create the proper DatabaseProperty type automaticaly.
    """

    __rel_map = {
        "title": DatabasePropertyTitle,
        "rich_text": DatabasePropertyText,
        "number": DatabasePropertyNumber,
        "select": DatabasePropertySelect,
        "multi_select": DatabasePropertyMultiSelect,
        "date": DatabasePropertyDate,
        "people": DatabasePropertyPeople,
        "files": DatabasePropertyFiles,
        "checkbox": DatabasePropertyCheckbox,
        "url": DatabasePropertyUrl,
        "email": DatabasePropertyEmail,
        "phone_number": DatabasePropertyPhoneNumber,
        "formula": DatabasePropertyFormula,
        "relation": DatabasePropertyRelation,
        "rollup": DatabasePropertyRollup,
        "created_by": DatabasePropertyCreatedBy,
        "created_time": DatabasePropertyCreatedTime,
        "last_edited_time": DatabasePropertyLastEditedTime,
        "last_edited_by": DatabasePropertyLastEditedBy,
    }

    @staticmethod
    def new(data):
        data_type = data.get("type")
        if data_type and data_type in DatabasePropertyFactory.__rel_map:
            return from_dict(
                data_class=DatabasePropertyFactory.__rel_map[data_type], data=data
            )
        else:
            raise Exception(f'DatabaseProperty "{data_type}" not implemented yet.')


@dataclass
class Database(__BaseDataClass):
    title: List[RichText]
    properties: object
    parent: Parent
    # icon: Optional[Emoji] = None  # TODO: Handle File
    cover: Optional[File] = None
    created_time: str = None
    last_edited_time: str = None
    url: str = None
    object: str = "database"
    id: str = None

    """
    __post_init__ is called right after the object is created.

    We use it here to keep self.properties as a Dictionnary but we
    want to convert values as Data Classes instances.
    """

    def __post_init__(self):
        for k in self.properties:
            self.properties[k] = DatabasePropertyFactory.new(self.properties[k])

    def __repr__(self):
        try:
            display(self.df())  # noqa: F821
            return ""
        except:  # noqa: E722
            return str(self)

    def schema(self):
        copied = deepcopy(self.properties)
        d = []
        for i in copied:
            p = copied[i]
            d.append({"Name": i, "Type": p.type})
        return pd.DataFrame(self.notion.to_dict(d))

    @classmethod
    def new(cls, title: str, page_id: str = None):
        parent = None
        if page_id:
            parent = Parent.new_page_parent(page_id=page_id)
        else:
            parent = Parent.new_workspace_parent()

        new_db = cls(title=[RichText.new(title)], parent=parent, properties={})
        new_db.add_property("Name", "title")
        return new_db

    @classmethod
    def from_dict(cls, data):
        return from_dict(data_class=cls, data=data)

    # def set_emoji_icon(
    #    self, data
    # ):  # TODO: Fix, seems like there is an issue with notion-client / httpx maybe.
    #    self.icon = Emoji(data)

    def duplicate(self):
        new_db = self.create()
        pages = self.query()
        for page in pages:
            page.parent.database_id = new_db.id
            page.duplicate()

        return new_db

    def add_property(self, col_name, type_name):
        data = {"type": type_name, "name": col_name, "id": ""}
        self.properties[col_name] = DatabasePropertyFactory.new(data)

    def df(self):
        pages = self.query({"page_size": 10})
        frames = []

        for page in pages:
            frame = {}
            for prop in page.properties:
                frame[prop] = str(page.properties[prop])
            frames.append(frame)

        return pd.DataFrame(frames)

    def query(self, query={}):
        return self.notion.databases.query(self.id, query)

    def update(self):
        return self.notion.databases.update(self)

    def create(self):
        return self.notion.databases.create(self)


"""
Page and Page properties
"""


@dataclass
class PageProperty(Property):
    pass


@dataclass
class PagePropertyTitle(PageProperty):
    title: Optional[List[RichText]] = field(default_factory=list)
    type: str = "title"

    @classmethod
    def new(cls, content: str) -> "PagePropertyTitle":
        return cls(title=[RichText.new(content)])

    def set_text(self, content: str):
        self.title = [RichText.new_text(content)]

    def __str__(self):
        return " ".join([v.plain_text for v in self.title])


@dataclass
class PagePropertyText(PageProperty):
    rich_text: Optional[List[RichText]] = field(default_factory=list)
    type: str = "rich_text"

    @classmethod
    def new(cls, content: str) -> "PagePropertyText":
        return cls(rich_text=[RichText.new(content)])

    def __str__(self):
        return " ".join([v.plain_text for v in self.rich_text])


@dataclass
class PagePropertyNumber(PageProperty):
    number: Optional[Number] = None
    type: str = "number"

    @classmethod
    def new(cls, number: Number) -> "PagePropertyNumber":
        return cls(number=number)

    def __str__(self):
        return str(self.number)


@dataclass
class PagePropertySelect_Value:
    name: str
    id: str = None
    color: str = "default"

    @classmethod
    def new(cls, name: str, color: str = "default") -> "PagePropertySelect_Value":
        return cls(name=name, color=color)


@dataclass
class PagePropertySelect(PageProperty):
    select: Optional[PagePropertySelect_Value] = field(
        default_factory=PagePropertySelect_Value
    )
    type: str = "select"

    @classmethod
    def new(cls, select: str, color: str = "default") -> "PagePropertySelect":
        return cls(select=PagePropertySelect_Value.new(name=select, color=color))

    def __str__(self):
        return str(self.select.name) if self.select else ""


@dataclass
class PagePropertyMultiSelect(PageProperty):
    multi_select: Optional[List[PagePropertySelect_Value]] = field(
        default_factory=list
    )  # TODO: Create own Value type.
    type: str = "multi_select"

    @classmethod
    def new(
        cls, values: List[str], color: str = "default"
    ) -> "PagePropertyMultiSelect":
        return cls(
            multi_select=[PagePropertySelect_Value.new(v, color) for v in values]
        )

    def __str__(self):
        return ", ".join([v.name for v in self.multi_select])


@dataclass
class PagePropertyDate_Value:
    start: str = None
    end: Optional[str] = None

    @classmethod
    def new(cls, start=str) -> "PagePropertyDate_Value":
        return cls(start=start)


@dataclass
class PagePropertyDate(PageProperty):
    date: Optional[PagePropertyDate_Value] = field(
        default_factory=PagePropertyDate_Value
    )
    type: str = "date"

    @classmethod
    def new(cls, date: str) -> "PagePropertyDate":
        return cls(date=PagePropertyDate_Value.new(date))

    def __str__(self):
        return str(self.date.start) if self.date else ""


@dataclass
class PagePropertyFormula(PageProperty):
    string: Optional[str] = None
    number: Optional[Number] = None
    boolean: Optional[bool] = None
    date: Optional[PagePropertyDate_Value] = field(
        default_factory=PagePropertyDate_Value
    )
    type: str = "formula"

    # TODO: classmethod new

    # TODO: classmethod new


@dataclass
class PagePropertyRelation(PageProperty):
    relation: List[object] = field(default_factory=list)
    type: str = "relation"

    # TODO: classmethod new


@dataclass
class PagePropertyRollup(PageProperty):
    number: Optional[Number] = None
    date: Optional[PagePropertyDate_Value] = field(
        default_factory=PagePropertyDate_Value
    )
    array: Optional[List[object]] = field(default_factory=list)
    type: str = "rollup"

    # TODO: classmethod new


@dataclass
class PagePropertyPeople(PageProperty):
    people: Optional[List[User]] = field(default_factory=list)
    type: str = "people"

    @classmethod
    def new(cls, people: List[str]) -> "PagePropertyPeople":
        return cls(people=[User.new(v) for v in people])

    def __str__(self):
        return " ".join([v.name for v in self.people])


@dataclass
class PagePropertyFiles(PageProperty):
    files: Optional[List[File]] = field(default_factory=list)
    type: str = "files"

    @classmethod
    def new(cls, files: List[str]) -> "PagePropertyFiles":
        return cls(files=[File.new(v) for v in files])

    def __str__(self):
        return ", ".join([v.name for v in self.files])


@dataclass
class PagePropertyCheckbox(PageProperty):
    checkbox: bool = None
    type: str = "checkbox"

    @classmethod
    def new(cls, checkbox: bool) -> "PagePropertyCheckbox":
        return cls(checkbox=checkbox)

    def __str__(self):
        return str(self.checkbox)


@dataclass
class PagePropertyUrl(PageProperty):
    url: Optional[str] = None
    type: str = "url"

    @classmethod
    def new(cls, url: str) -> "PagePropertyUrl":
        return cls(url=url)

    def __str__(self):
        return str(self.url)


@dataclass
class PagePropertyEmail(PageProperty):
    email: Optional[str] = None
    type: str = "email"

    @classmethod
    def new(cls, email: str) -> "PagePropertyEmail":
        return cls(email=email)

    def __str__(self):
        return str(self.email)


@dataclass
class PagePropertyPhoneNumber(PageProperty):
    phone_number: Optional[str] = None
    type: str = "phone_number"

    @classmethod
    def new(cls, phone_number: str) -> "PagePropertyPhoneNumber":
        return cls(phone_number=phone_number)

    def __str__(self):
        return str(self.phone_number)


@dataclass
class PagePropertyCreatedBy(PageProperty):
    created_by: User = field(default_factory=User)
    type: str = "created_by"

    def __str__(self):
        return str(self.created_by.name) if self.created_by else ""


@dataclass
class PagePropertyCreatedTime(PageProperty):
    created_time: str = None
    type: str = "created_time"

    def __str__(self):
        return str(self.created_time)


@dataclass
class PagePropertyLastEditedTime(PageProperty):
    last_edited_time: str = None
    type: str = "last_edited_time"

    def __str__(self):
        return str(self.last_edited_time)


@dataclass
class PagePropertyLastEditedBy(PageProperty):
    last_edited_by: User = field(default_factory=User)
    type: str = "last_edited_by"

    def __str__(self):
        return str(self.last_edited_by.name) if self.last_edited_by else ""


class PagePropertyFactory:
    """
    This class is a helper to create the proper PageProperty type automaticaly.
    """

    __rel_map = {
        "title": PagePropertyTitle,
        "rich_text": PagePropertyText,
        "number": PagePropertyNumber,
        "select": PagePropertySelect,
        "multi_select": PagePropertyMultiSelect,
        "date": PagePropertyDate,
        "people": PagePropertyPeople,
        "files": PagePropertyFiles,
        "checkbox": PagePropertyCheckbox,
        "url": PagePropertyUrl,
        "email": PagePropertyEmail,
        "phone_number": PagePropertyPhoneNumber,
        "formula": PagePropertyFormula,
        "relation": PagePropertyRelation,
        "rollup": PagePropertyRollup,
        "created_by": PagePropertyCreatedBy,
        "created_time": PagePropertyCreatedTime,
        "last_edited_time": PagePropertyLastEditedTime,
        "last_edited_by": PagePropertyLastEditedBy,
    }

    @staticmethod
    def new(data):
        data_type = data.get("type")
        if data_type and data_type in PagePropertyFactory.__rel_map:
            return from_dict(
                data_class=PagePropertyFactory.__rel_map[data_type], data=data
            )
        else:
            raise Exception(f'PageProperty "{data_type}" not implemented yet.')

    @staticmethod
    def new_default(type: str, payload=any, color=None):
        if type in ["select", "multi_select"]:
            return PagePropertyFactory.__rel_map[type].new(payload, color=color)
        else:
            return PagePropertyFactory.__rel_map[type].new(payload)


@dataclass
class Page(__BaseDataClass):
    properties: object
    parent: Parent
    blocks: Optional[List["Block"]] = field(default_factory=list)
    archived: bool = False
    # icon: Optional[Emoji] = None  # TODO: Handle File
    cover: Optional[File] = None
    created_time: str = None
    last_edited_time: str = None
    url: str = None
    object: str = "page"
    id: str = None

    """
    __post_init__ is called right after the object is created.

    We use it here to keep self.properties as a Dictionnary but we
    want to convert values as Data Classes instances.
    """

    def __post_init__(self):
        for k in self.properties:
            self.properties[k] = PagePropertyFactory.new(self.properties[k])

    def __repr__(self):
        try:
            display(self.df())  # noqa: F821
            return ""
        except:  # noqa: E722
            return str(self)

    def get_blocks(self):
        self.blocks = self.notion.blocks.children(self.id)
        return self.blocks

    def duplicate(self):
        page = deepcopy(self)
        page.id = None
        page.parent.type = None

        to_delete = []
        for col_name in page.properties:
            prop = page.properties[col_name]
            if prop.type in [
                "last_edited_time",
                "created_time",
                "created_by",
                "last_edited_by",
            ]:
                to_delete.append(col_name)
            elif getattr(prop, prop.type) is None:
                to_delete.append(col_name)

        for col_name in to_delete:
            del page.properties[col_name]

        block_tree = BlockTree(self.notion.blocks.get(self.id))
        new_page = page.create()
        block_tree.duplicate(new_page.id)

        return new_page

    def df(self, pivot=False):
        copied = deepcopy(self.properties)
        if pivot is False:
            d = []
            for i in copied:
                p = copied[i]

                d.append({"Name": i, "Type": p.type, "Value": str(p)})
            return pd.DataFrame(self.notion.to_dict(d))
        else:
            frame = {}
            for i in copied:
                p = copied[i]
                frame[i] = str(p)
            return pd.DataFrame(self.notion.to_dict([frame]))

    @classmethod
    def new(cls, page_id: str = None, database_id: str = None):
        parent = None
        if page_id:
            parent = Parent.new_page_parent(page_id=page_id)
        elif database_id:
            parent = Parent.new_database_parent(database_id=database_id)
        else:
            parent = Parent.new_workspace_parent()
        #  For some reasons, not explained in Notion API Documentation, when creating a new page
        # the property 'type' of a Parent object must be omitted.
        parent.type = None

        return cls(parent=parent, properties={})

    def add_block(self, block_type: str):
        new_block = Block.new(type=block_type)
        new_block = new_block.append_to(page_id=self.id)
        self.get_blocks()
        return new_block

    def update(self):
        ret = self.notion.pages.update(self)
        blocks_to_add = []
        for block in self.blocks:
            if not block.id:
                blocks_to_add.append(block)

        if len(blocks_to_add) > 0:
            self.notion.blocks.append(self.id, blocks_to_add)

        return ret

    def create(self):
        return self.notion.pages.create_from_page(self)

    """
    Properties setters
    """

    def __generic_property_setter(self, type: str, column_name, column_value):
        if column_name not in self.properties and self.parent.type == "database_id":
            db = self.notion.databases.retrieve(database_id=self.parent.database_id)
            db.add_property(column_name, type)
            db.update()
        if type in ["select", "multi_select"]:
            db = self.notion.databases.retrieve(database_id=self.parent.database_id)
            select_option = db.properties[column_name].get_option_by_name(column_value)
            color = None
            if select_option:
                color = select_option.color
            self.properties[column_name] = PagePropertyFactory.new_default(
                type, column_value, color=color
            )
        else:
            self.properties[column_name] = PagePropertyFactory.new_default(
                type, column_value
            )
        return self.properties[column_name]

    def title(self, *k):
        return self.__generic_property_setter("title", *k)

    def rich_text(self, *k):
        return self.__generic_property_setter("rich_text", *k)

    def number(self, *k):
        return self.__generic_property_setter("number", *k)

    def select(self, *k):
        return self.__generic_property_setter("select", *k)

    def multi_select(self, *k):
        return self.__generic_property_setter("multi_select", *k)

    def date(self, *k):
        return self.__generic_property_setter("date", *k)

    def people(self, *k):
        return self.__generic_property_setter("people", *k)

    def checkbox(self, *k):
        return self.__generic_property_setter("checkbox", *k)

    def link(self, *k):
        return self.__generic_property_setter("url", *k)

    def email(self, *k):
        return self.__generic_property_setter("email", *k)

    def phone_number(self, *k):
        return self.__generic_property_setter("phone_number", *k)

    """
    Block setters
    """

    def __generic_block_setter(self, type: str, payload):
        new_block = BlockTypeFactory.new_default(type, payload)
        self.blocks.append(new_block)
        return new_block

    def paragraph(self, *k):
        return self.__generic_block_setter("paragraph", *k)

    def heading_1(self, *k):
        return self.__generic_block_setter("heading_1", *k)

    def heading_2(self, *k):
        return self.__generic_block_setter("heading_2", *k)

    def heading_3(self, *k):
        return self.__generic_block_setter("heading_3", *k)

    def bulleted_list_item(self, *k):
        return self.__generic_block_setter("bulleted_list_item", *k)

    def numbered_list_item(self, *k):
        return self.__generic_block_setter("numbered_list_item", *k)

    def to_do(self, *k):
        return self.__generic_block_setter("to_do", *k)

    def toggle(self, *k):
        return self.__generic_block_setter("toggle", *k)

    def code(self, *k):
        return self.__generic_block_setter("code", *k)

    def embed(self, *k):
        return self.__generic_block_setter("embed", *k)

    def image(self, *k):
        return self.__generic_block_setter("image", *k)

    def video(self, *k):
        return self.__generic_block_setter("video", *k)

    def file(self, *k):
        return self.__generic_block_setter("file", *k)

    def pdf(self, *k):
        return self.__generic_block_setter("pdf", *k)

    def bookmark(self, *k):
        return self.__generic_block_setter("bookmark", *k)

    def equation(self, *k):
        return self.__generic_block_setter("equation", *k)

    def divider(self, *k):
        return self.__generic_block_setter("divider", *k)

    def table_of_contents(self, *k):
        return self.__generic_block_setter("table_of_contents", *k)

    def child_database(self, *k):
        return self.__generic_block_setter("child_database", *k)


"""
Block and Blocks properties.
"""


@dataclass
class Block(__BaseDataClass):
    type: str
    paragraph: Optional["BlockParagraph"] = None
    heading_1: Optional["BlockHeadingOne"] = None
    heading_2: Optional["BlockHeadingTwo"] = None
    heading_3: Optional["BlockHeadingThree"] = None
    callout: Optional["BlockCallout"] = None
    quote: Optional["BlockQuote"] = None
    bulleted_list_item: Optional["BlockBulletedListItem"] = None
    numbered_list_item: Optional["BlockNumberedListItem"] = None
    to_do: Optional["BlockToDo"] = None
    toggle: Optional["BlockToggle"] = None
    code: Optional["BlockCode"] = None
    child_page: Optional["BlockChildPage"] = None
    child_database: Optional["BlockChildDatabase"] = None
    embed: Optional["BlockEmbed"] = None
    image: Optional["File"] = None
    video: Optional["File"] = None
    file: Optional["File"] = None
    pdf: Optional["File"] = None
    bookmark: Optional["BlockBookmark"] = None
    equation: Optional["BlockEquation"] = None
    divider: Optional[dict] = None
    table_of_contents: Optional[dict] = None
    has_children: bool = False
    archived: bool = False
    created_time: str = None
    last_edited_time: str = None
    object: str = "block"
    id: str = None

    @classmethod
    def new(cls, type: str, prop: any = None):
        if prop:
            return from_dict(data_class=cls, data={"type": type, type: prop})
        return from_dict(data_class=cls, data={"type": type, type: {}})

    # def __post_init__(self):
    #     if self.has_children is True:
    #         self.children = self.notion.blocks.children(block_id=self.id)

    # TODO: To delete
    # def __post_init__(self):
    #   setattr(self, self.type, BlockTypeFactory.new(self))

    def append_to(self, page_id):
        return self.notion.blocks.append(page_id, self)

    def update(self):
        return self.notion.blocks.update(self)

    def delete(self):
        return self.notion.blocks.delete(self.id)


@dataclass
class BlockParagraph(__BaseDataClass):
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    def add(self, content):
        self.text.append(RichText.new_text(content=content))

    @classmethod
    def new(cls, content: str) -> "BlockParagraph":
        return cls(children=[], text=[RichText.new(content)])


@dataclass
class BlockHeadingOne(__BaseDataClass):
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockHeadingOne":
        return cls(text=[RichText.new(content)])


@dataclass
class BlockHeadingTwo(__BaseDataClass):
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockHeadingTwo":
        return cls(text=[RichText.new(content)])


@dataclass
class BlockHeadingThree(__BaseDataClass):
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockHeadingThree":
        return cls(text=[RichText.new(content)])


@dataclass
class BlockCallout(__BaseDataClass):
    icon: Optional[Union[File, Emoji]]
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockCallout":
        return cls(icon=None, children=[], text=[RichText.new(content)])


@dataclass
class BlockQuote(__BaseDataClass):
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockQuote":
        return cls(children=[], text=[RichText.new(content)])


@dataclass
class BlockBulletedListItem(__BaseDataClass):
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockBulletedListItem":
        return cls(children=[], text=[RichText.new(content)])


@dataclass
class BlockNumberedListItem(__BaseDataClass):
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockNumberedListItem":
        return cls(children=[], text=[RichText.new(content)])


@dataclass
class BlockToDo(__BaseDataClass):
    checked: Optional[bool]
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockToDo":
        return cls(checked=False, children=[], text=[RichText.new(content)])


@dataclass
class BlockToggle(__BaseDataClass):
    children: Optional[List[Block]]
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockToggle":
        return cls(children=[], text=[RichText.new(content)])


@dataclass
class BlockCode(__BaseDataClass):
    language: Optional[str] = "plain text"
    text: List[RichText] = field(default_factory=list)

    @classmethod
    def new(cls, content: str) -> "BlockCode":
        return cls(text=[RichText.new(content)])


@dataclass
class BlockChildPage(__BaseDataClass):
    title: Optional[str] = "New page"


@dataclass
class BlockChildDatabase(__BaseDataClass):
    title: Optional[str] = ""

    @classmethod
    def new(cls, title: str) -> "BlockChildDatabase":
        return cls(title=title)


@dataclass
class BlockEmbed(__BaseDataClass):
    url: Optional[str] = ""

    @classmethod
    def new(cls, url: str) -> "BlockEmbed":
        return cls(url=url)


@dataclass
class BlockImage(__BaseDataClass):
    image: Optional[File] = field(default_factory=File)

    @classmethod
    def new(cls, url: str) -> "BlockImage":
        return cls(image=File.new(url))


@dataclass
class BlockVideo(__BaseDataClass):
    video: Optional[File] = field(default_factory=File)

    @classmethod
    def new(cls, url: str) -> "BlockVideo":
        return cls(video=File.new(url))


@dataclass
class BlockFile(__BaseDataClass):
    file: Optional[File] = field(default_factory=File)

    @classmethod
    def new(cls, url: str) -> "BlockFile":
        return cls(file=File.new(url))


@dataclass
class BlockPdf(__BaseDataClass):
    pdf: Optional[File] = field(default_factory=File)

    @classmethod
    def new(cls, url: str) -> "BlockPdf":
        return cls(pdf=File.new(url))


@dataclass
class BlockBookmark(__BaseDataClass):
    caption: List[RichText] = field(default_factory=list)
    url: str = ""

    @classmethod
    def new(cls, content: str) -> "BlockBookmark":
        return cls(caption=[RichText.new(content)])


@dataclass
class BlockEquation(__BaseDataClass):
    expression: str = ""

    @classmethod
    def new(cls, expression: str) -> "BlockEquation":
        return cls(expression=expression)


@dataclass
class BlockDivider(__BaseDataClass):
    divider: object = field(default_factory=dict)

    @classmethod
    def new(cls) -> "BlockDivider":
        return cls()


@dataclass
class BlockTableOfContents(__BaseDataClass):
    table_of_contents: object = field(default_factory=dict)

    @classmethod
    def new(cls) -> "BlockTableOfContents":
        return cls()


class BlockTypes(Enum):
    paragraph = "paragraph"
    heading_1 = "heading_1"
    heading_2 = "heading_2"
    heading_3 = "heading_3"
    callout = "callout"
    quote = "quote"
    bulleted_list_item = "bulleted_list_item"
    numbered_list_item = "numbered_list_item"
    to_do = "to_do"
    toggle = "toggle"
    code = "code"
    child_page = "child_page"
    child_database = "child_database"
    embed = "embed"
    image = "image"
    video = "video"
    file = "file"
    pdf = "pdf"
    bookmark = "bookmark"
    equation = "equation"
    divider = "divider"
    table_of_contents = "table_of_contents"
    unsupported = "unsupported"


class BlockTypeFactory:
    """
    This class is a helper to create the proper BlockType automaticaly.
    """

    __rel_map = {
        "paragraph": BlockParagraph,
        "heading_1": BlockHeadingOne,
        "heading_2": BlockHeadingTwo,
        "heading_3": BlockHeadingThree,
        "callout": BlockCallout,  # Seems to be missing documentation as of 21/10/17
        "quote": BlockQuote,  # Seems to be missing documentation as of 21/10/17
        "bulleted_list_item": BlockBulletedListItem,
        "numbered_list_item": BlockNumberedListItem,
        "to_do": BlockToDo,
        "toggle": BlockToggle,
        "code": BlockCode,  # Seems to be missing documentation as of 21/10/17
        "child_page": BlockChildPage,
        "child_database": BlockChildDatabase,
        "embed": BlockEmbed,
        "image": File,
        # "video": BlockVideo,
        "video": File,
        "file": File,
        "pdf": File,
        "bookmark": BlockBookmark,
        "equation": BlockEquation,  # Seems to be missing documentation as of 21/10/17
        "divider": dict,  # Seems to be missing documentation as of 21/10/17
        "table_of_contents": dict,  # Seems to be missing documentation as of 21/10/17
        # "unsupported": BlockUnsupported
    }

    # @staticmethod
    # def new(data):
    #    data_type = getattr(data, 'type')
    #    if data_type and data_type in BlockTypeFactory.__rel_map:
    #        return from_dict(data_class=BlockTypeFactory.__rel_map[data_type], data=getattr(data, data_type))
    #    else:
    #        raise Exception(f'BlockType "{data_type}" not implemented yet.')

    @staticmethod
    def new_default(type: str, payload: any):
        if type in ["divider", "table_of_contents"]:
            prop = BlockTypeFactory.__rel_map[type].new()
        else:
            prop = BlockTypeFactory.__rel_map[type].new(payload)
        return Block.new(type, prop)


"""Dataclass json encoder"""


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)


def del_none(d):
    """
    Delete keys with the value ``None`` in a dictionary, recursively.

    This alters the input so you may wish to ``copy`` the dict first.
    """
    # For Python 3, write `list(d.items())`; `d.items()` won’t work
    # For Python 2, write `d.items()`; `d.iteritems()` won’t work
    if type(d) is dict:
        for key, value in list(d.items()):
            if value is None:
                del d[key]
            elif isinstance(value, dict) or isinstance(value, list):
                del_none(value)
    elif type(d) is list:
        for e in d:
            del_none(e)
    return d  # For convenience


class BlockTree:
    block: Block = None
    children: List["BlockTree"] = None

    def __init__(self, block: Block):
        self.block = block
        if self.block.has_children is True:
            self.children = []
            blocks = Notion.instance().blocks.children(self.block.id)
            for b in blocks:
                self.children.append(BlockTree(b))

    def copy(self):
        copied_block = deepcopy(self.block)
        copied_block.id = None
        copied_block.created_time = None
        copied_block.last_edited_time = None
        return copied_block

    def duplicate_children(self, parent_id):
        if self.children is not None:
            to_append = []
            for child in self.children:
                to_append.append(child.copy())
            new_blocks = Notion.instance().block.append(parent_id, to_append)
            count = 0
            for child in self.children:
                child.duplicate_children(new_blocks[count].id)
                count = count + 1

    def duplicate(self, parent_id):
        if self.children is not None:
            to_append = []
            for child in self.children:
                to_append.append(child.copy())
            new_blocks = Notion.instance().block.append(parent_id, to_append)

            threads = []
            count = 0
            for child in self.children:
                x = threading.Thread(
                    target=child.duplicate_children, args=(new_blocks[count].id,)
                )
                threads.append(x)
                x.start()
                count = count + 1

            for t in threads:
                t.join()



import pandas as pd
import requests
from datetime import datetime
import plotly.graph_objects as go
import urllib
from naas_drivers.tools.emailbuilder import EmailBuilder
import pydash as _pd

emailbuilder = EmailBuilder()

QONTO_API_URL = "https://thirdparty.qonto.com/v2"
DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
NUMBER_FORMAT = "{:,.2f} €"

QONTO_CARD = "https://lib.umso.co/lib_sluGpRGQOLtkyEpz/98a28jf4yswpuert.png"
NAAS_WEBSITE = "https://www.naas.ai"


class Qonto:
    @staticmethod
    def get_dates(df, date_column, date_from=None, date_to=None):
        # Get all dates for range
        filter_df = 0
        dates = []
        if date_to is None:
            date_to = df[date_column].max()
        else:
            date_to = datetime.strptime(date_to, DATE_FORMAT)
        if date_from is None:
            date_from = df[date_column].min()
        elif type(date_from) == int:
            filter_df = date_from
            date_from = df[date_column].min()
        else:
            date_from = datetime.strptime(date_from, DATE_FORMAT)
        dates_range = pd.date_range(start=date_from, end=date_to)
        for date in dates_range:
            date = str(date.strftime(DATE_FORMAT))
            dates.append(date)
        return dates[filter_df:]

    @staticmethod
    def filter_dates(df, date_column, date_from=None, date_to=None):
        # Create columns date temp to apply filter
        df["DATE_TMP"] = pd.to_datetime(df[date_column]).dt.strftime(DATE_FORMAT)

        # Get list of dates
        dates = Qonto.get_dates(df, "DATE_TMP", date_from=date_from, date_to=date_to)

        # Filter on new columns
        df = df[df["DATE_TMP"].isin(dates)].reset_index(drop=True)
        return df.drop("DATE_TMP", axis=1)

    def connect(self, user_id, api_token):
        # Init thinkific attribute
        self.user_id = user_id
        self.api_token = api_token

        # Init headers
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"{self.user_id}:{self.api_token}",
        }

        # Init end point
        self.organizations = Organizations(self.user_id, self.headers)
        self.positions = Organizations(self.user_id, self.headers)
        self.transactions = Transactions(self.user_id, self.headers)
        self.statements = Statements(self.user_id, self.headers)

        # Set connexion to active
        self.connected = True
        return self


class Organizations(Qonto):
    def __init__(self, user_id, headers):
        Qonto.__init__(self)
        self.user_id = user_id
        self.headers = headers

    def get(self, cols_to_drop=["SLUG", "BALANCE_CENTS", "AUTHORIZED_BALANCE_CENTS"]):
        """
        Return an dataframe object with 11 columns:
        - IBAN
        - BIC
        - CURRENCY
        - BALANCE
        - AUTHORIZED_BALANCE
        - NAME
        - UPDATED_AT
        - STATUS
        - ORGANIZATION_SLUG
        - LEGAL_NAME
        - EXTRACT_DATE

        Parameters
        ----------
        cols_to_drop: list (default None):
            Columns to drop from your dataframe.
        """

        req_url = f"{QONTO_API_URL}/organization"
        res = requests.get(req_url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        res_json = res.json()

        # Get bank accounts
        slug = _pd.get(res_json, "organization.slug")
        legal_name = _pd.get(res_json, "organization.legal_name")
        accounts = _pd.get(res_json, "organization.bank_accounts")
        df = pd.DataFrame.from_records(accounts)
        df["organization_slug"] = slug
        df["legal_name"] = legal_name

        # Cleaning naas
        df["extract_date"] = datetime.now().strftime(DATETIME_FORMAT)
        df = df.fillna("Not defined")
        df.columns = df.columns.str.upper()

        # Columns to drop
        for cols in cols_to_drop:
            df = df.drop(cols, axis=1)
        return df


class Transactions(Qonto):
    def __init__(self, user_id, headers):
        Qonto.__init__(self)
        self.user_id = user_id
        self.headers = headers

    def get(self, date="EMITTED_AT", date_from=None, date_to=None):
        """
        Return an dataframe object with 22 columns:
        - IBAN
        - SETTLED_AT
        - EMITTED_AT
        - TRANSACTION_ID
        - TRANSACTION_ORDER
        - LABEL
        - STATUS
        - CATEGORY
        - REFERENCE
        - OPERATION_TYPE
        - CARD_LAST_DIGITS
        - SIDE
        - AMOUNT
        - CURRENCY
        - VAT_AMOUNT
        - VAT_RATE
        - LOCAL_AMOUNT
        - LOCAL_CURRENCY
        - NOTE
        - ATTACHMENT_IDS
        - ATTACHMENT_LOST
        - ATTACHMENT_REQUIRED

        Parameters
        ----------
        date: str (default "EMITTED_AT"):
            Date column to filter on.
        date_from: date (default None):
            Date from to get data, format "%Y-%m-%d".
        date_to: date (default None):
            Date to to get data, format "%Y-%m-%d".
        """
        # Get organizations
        df_organisations = Organizations.get(self)

        # For each bank account, get all transactions
        df_transaction = pd.DataFrame()
        for _, row in df_organisations.iterrows():
            iban = row["IBAN"]

            # Get transactions
            has_more = True
            current_page = 1
            while has_more:
                params = {
                    "current_page": current_page,
                    "iban": iban,
                }
                res = requests.get(
                    url=f"{QONTO_API_URL}/transactions?{urllib.parse.urlencode(params, safe='(),')}",
                    headers=self.headers,
                )
                try:
                    res.raise_for_status()
                except requests.HTTPError as e:
                    return e
                items = res.json()
                transactions = items.get("transactions")
                df = pd.DataFrame.from_records(transactions)
                df["iban"] = iban
                df_transaction = pd.concat([df_transaction, df], axis=0)
                # Check if next page exists
                next_page = items.get("meta").get("next_page")
                if next_page is None:
                    has_more = False
                else:
                    current_page = int(next_page)

        # Formatting
        df_transaction["transaction_order"] = df_transaction.apply(
            lambda row: int(row["transaction_id"].split("-")[-1]), axis=1
        )
        df_transaction = df_transaction.sort_values(
            by=["iban", "transaction_order"]
        ).reset_index(drop=True)

        # Select columns to keep
        to_keep = [
            "iban",
            "settled_at",
            "emitted_at",
            "transaction_id",
            "transaction_order",
            "label",
            "status",
            "category",
            "reference",
            "operation_type",
            "card_last_digits",
            "side",
            "amount",
            "currency",
            "vat_amount",
            "vat_rate",
            "local_amount",
            "local_currency",
            "note",
            "attachment_ids",
            "attachment_lost",
            "attachment_required",
        ]
        df_transaction = df_transaction[to_keep]

        # Sign amounts
        df_transaction.loc[
            df_transaction["side"] == "debit", "amount"
        ] = df_transaction["amount"] * (-1)
        df_transaction.loc[
            df_transaction["side"] == "debit", "local_amount"
        ] = df_transaction["local_amount"] * (-1)
        df_transaction.loc[
            df_transaction["side"] == "debit", "vat_amount"
        ] = df_transaction["vat_amount"] * (-1)

        # Format columns in upper case
        df_transaction = df_transaction.fillna("Not defined")
        df_transaction.columns = df_transaction.columns.str.upper()

        # Filter dataframe
        df_transaction = Qonto.filter_dates(
            df_transaction, date, date_from=date_from, date_to=date_to
        )
        return df_transaction


class Statements(Transactions):
    def __init__(self, user_id, headers):
        Qonto.__init__(self)
        self.user_id = user_id
        self.headers = headers

    def get(
        self,
        date="EMITTED_AT",
        to_group=[
            "IBAN",
            "DATE",
            "TRANSACTION_ID",
            "TRANSACTION_ORDER",
            "LABEL",
            "REFERENCE",
            "CATEGORY",
            "OPERATION_TYPE",
            "CURRENCY",
        ],
        date_from=None,
        date_to=None,
    ):
        """
        Return an dataframe object with 11 columns:
        - IBAN
        - DATE
        - TRANSACTION_ID
        - TRANSACTION_ORDER
        - LABEL
        - REFERENCE
        - CATEGORY
        - OPERATION_TYPE
        - CURRENCY
        - AMOUNT
        - POSITION

        Parameters
        ----------
        date: str (default "EMITTED_AT"):
            Date column to filter on.
        date_from: date (default None):
            Date from to get data, format "%Y-%m-%d".
        date_to: date (default None):
            Date to to get data, format "%Y-%m-%d".
        """

        # Get transactions
        df = Transactions.get(self, date)

        # Set date column
        df = df.rename(columns={date: "DATE"})
        df["DATE"] = pd.to_datetime(df["DATE"]).dt.strftime(DATE_FORMAT)

        # Groupby
        if "DATE" not in to_group:
            to_group = ["DATE"] + to_group
        if "IBAN" not in to_group:
            to_group = ["IBAN"] + to_group
        df = df.groupby(to_group, as_index=False).agg({"AMOUNT": "sum"})

        # Calc position
        if "TRANSACTION_ORDER" in to_group:
            df = df.sort_values(by=["IBAN", "TRANSACTION_ORDER"]).reset_index(drop=True)
        else:
            df = df.sort_values(by=["IBAN", "DATE"]).reset_index(drop=True)
        df_statement = pd.DataFrame()
        ibans = df["IBAN"].unique()
        for iban in ibans:
            tmp = df[df["IBAN"] == iban]
            tmp["POSITION"] = tmp.agg({"AMOUNT": "cumsum"})
            df_statement = pd.concat([df_statement, tmp])

        # Filter dataframe
        df_statement = Qonto.filter_dates(
            df_statement, date_column="DATE", date_from=date_from, date_to=date_to
        )
        return df_statement

    def barline(
        self,
        to_group=["IBAN", "DATE"],
        date_from=None,
        date_to=None,
        title=None,
        line_name="Solde",
        line_color="#1ea1f1",
        cashin_name="Encaissements",
        cashin_color="#47dd82",
        cashout_name="Décaissements",
        cashout_color="#ea484f",
    ):
        # Data linechart
        df = self.get(to_group=to_group, date_from=date_from, date_to=date_to)
        df_line = df.copy()
        df_line = df_line[["DATE", "POSITION"]]
        df_line["DATE"] = pd.to_datetime(df_line["DATE"])

        # Data bar
        df_bar = self.get(
            to_group=["TRANSACTION_ID", "LABEL", "OPERATION_TYPE"],
            date_from=date_from,
            date_to=date_to,
        )
        df_bar.loc[df_bar.AMOUNT > 0, "FLOWS"] = "CASH_IN"
        df_bar.loc[df_bar.AMOUNT < 0, "FLOWS"] = "CASH_OUT"
        df_bar["YEAR"] = pd.to_datetime(df_bar["DATE"]).dt.strftime("%Y")
        df_bar["MONTH"] = pd.to_datetime(df_bar["DATE"]).dt.strftime("%m")
        df_bar["DATE"] = df_bar["YEAR"] + "-" + df_bar["MONTH"]
        df_bar = df_bar.groupby(["DATE", "FLOWS"], as_index=False).agg(
            {"AMOUNT": "sum"}
        )
        cashin = df_bar[df_bar["FLOWS"] == "CASH_IN"]
        cashout = df_bar[df_bar["FLOWS"] == "CASH_OUT"]

        # Title
        if title is None:
            # Get date from and date to if None
            if date_from is None:
                date_from = df.DATE.unique().min()
            if date_to is None:
                date_to = df.DATE.unique().max()
            # Format date
            date_from = datetime.strptime(date_from, DATE_FORMAT).strftime("%d/%m/%Y")
            date_to = datetime.strptime(date_to, DATE_FORMAT).strftime("%d/%m/%Y")
            # Create title
            title = f"Evolution du {date_from} au {date_to}"
        # Init fig
        fig = go.Figure()

        # Create chart
        fig.add_trace(
            go.Scatter(
                x=df_line["DATE"],
                y=df_line["POSITION"],
                name=line_name,
                marker=dict(color=line_color),
            )
        )
        fig.add_bar(
            x=cashin["DATE"],
            y=cashin["AMOUNT"],
            name=cashin_name,
            marker=dict(color=cashin_color),
        )
        fig.add_bar(
            x=cashout["DATE"],
            y=cashout["AMOUNT"],
            name=cashout_name,
            marker=dict(color=cashout_color),
        )

        # Update chart display
        fig.update_layout(
            plot_bgcolor="#ffffff",
            hovermode="x",
            title=title,
            title_font=dict(family="Arial", size=18, color="black"),
            width=1200,
            height=800,
            margin_pad=10,
        )
        fig.update_yaxes(tickprefix="€", gridcolor="#eaeaea")
        fig.update_xaxes(dtick="M1")
        return fig

    def summary(self, summary_type, language="EN", date_from=None, date_to=None):
        # Get data
        df_statement = self.get(
            to_group=[summary_type], date_from=date_from, date_to=date_to
        )

        # Create
        if date_from is None:
            date_from = df_statement.DATE.unique().min()
        if date_to is None:
            date_to = df_statement.DATE.unique().max()

        # > Get first position
        first_position = round(df_statement["POSITION"].tolist()[0], 2)
        first_text = f"{datetime.strptime(date_from, DATE_FORMAT).strftime('%d/%m/%Y')}"
        df_first = pd.DataFrame(
            [{summary_type: first_text, "ORDER": 0, "AMOUNT": first_position}]
        )
        # > Get last position
        last_position = round(df_statement["POSITION"].tolist()[-1], 2)
        current_text = f"{datetime.strptime(date_to, DATE_FORMAT).strftime('%d/%m/%Y')}"
        df_last = pd.DataFrame(
            [{summary_type: current_text, "ORDER": 99999, "AMOUNT": last_position}]
        )

        # > Get transactions
        df_transaction = df_statement.copy()

        # Order
        if summary_type == "OPERATION_TYPE":
            # Order
            transaction_order = {
                "income": 1,
                "swift_income": 2,
                "card": 3,
                "transfer": 4,
                "qonto_fee": 5,
                "direct_debit": 6,
                "cheque": 7,
                "recall": 8,
            }
            df_transaction["ORDER"] = df_transaction[summary_type].replace(
                transaction_order
            )
        else:
            values = sorted(df_transaction[summary_type].unique())
            transaction_order = {}
            for i, v in enumerate(values):
                transaction_order[v] = i
            df_transaction["ORDER"] = df_transaction[summary_type].replace(
                transaction_order
            )

        # Label
        if language == "FR":
            # Replace value
            transaction_type = {
                "income": " - Encaissements",
                "swift_income": " - SWIFT",
                "card": " - Flux carte bleue",
                "transfer": " - Virements",
                "qonto_fee": " - Frais bancaires",
                "direct_debit": " - Prélèvements",
                "cheque": " - Chèques",
                "recall": " -  Rappel",
            }
            df_transaction[summary_type] = df_transaction[summary_type].replace(
                transaction_type
            )
            value_col = "Montant"
        else:
            df_transaction[summary_type] = " - " + df_transaction[
                summary_type
            ].str.capitalize().str.replace("_", " ")
            value_col = "Amount"

        # Groupby month
        to_group = [summary_type, "ORDER"]
        to_agg = {"AMOUNT": "sum"}
        df_transaction = df_transaction.groupby(to_group, as_index=False).agg(to_agg)

        # Create summary table
        cash_summary = pd.concat([df_first, df_transaction, df_last], axis=0)
        cash_summary = cash_summary.sort_values("ORDER").drop("ORDER", axis=1)
        cash_summary["AMOUNT"] = (
            cash_summary["AMOUNT"].map(NUMBER_FORMAT.format).str.replace(",", " ")
        )
        cash_summary = cash_summary.rename(
            columns={summary_type: "Type", "AMOUNT": value_col}
        )
        return cash_summary

    def transactions(self, date_from=None, date_to=None):
        # Data
        df = self.get(
            to_group=["LABEL", "OPERATION_TYPE"], date_from=date_from, date_to=date_to
        )

        # Groupby
        to_group = ["LABEL", "OPERATION_TYPE", "DATE"]
        df = df.groupby(to_group, as_index=False).agg({"AMOUNT": "sum"})
        df["DATE"] = pd.to_datetime(df["DATE"]).dt.strftime("%d/%m/%Y")
        # Replace value
        transaction_type = {
            "income": "Encaissement",
            "swift_income": "SWIFT",
            "card": "Carte bleue",
            "transfer": "Virement",
            "qonto_fee": "Frais bancaires",
            "direct_debit": "Prélèvement",
            "cheque": "Chèque",
            "recall": "Rappel",
        }
        df["OPERATION_TYPE"] = df["OPERATION_TYPE"].replace(transaction_type)

        # Rename columns
        week_cols = {
            "LABEL": "Description",
            "OPERATION_TYPE": "Type",
            "DATE": "Date",
            "AMOUNT": "Montant",
        }
        df = df.rename(columns=week_cols)
        df["Montant"] = (
            df["Montant"].map("{:,.2f} €".format).str.replace(",", " ")
        )  # .str.replace(".", ",")
        df = df.sort_values("Date", ascending=False)
        return df

    def email(
        self,
        date_from,
        date_to,
        last_position,
        graph_img,
        graph_link,
        cash_summary,
        nb_last,
        df_last,
        statement_link,
    ):
        # Format variable
        date_from = datetime.strptime(date_from, DATE_FORMAT).strftime("%d/%m/%Y")
        date_to = datetime.strptime(date_to, DATE_FORMAT).strftime("%d/%m/%Y")
        last_position = NUMBER_FORMAT.format(last_position).replace(
            ",", " "
        )  # .replace(".", ",")

        # Table style
        table_style = {
            "border": True,
            "header": True,
            "header_bg_color": "#6b5aed",
            "header_ft_color": "white",
            #             'border_color': '#BFBFBF',
        }

        table_cash = {
            "col_size": {
                0: "80%",
                1: "20%",
            },
            "col_align": {
                0: "left",
                1: "right",
            },
        }

        table_last_week = {
            "col_size": {0: "40%", 1: "20%", 2: "20%", 3: "20%"},
            "col_align": {
                0: "left",
                1: "center",
                2: "center",
                3: "right",
            },
        }

        # Email content
        content = {
            "title": (
                f"<a href='{NAAS_WEBSITE}'>"
                f"<img align='center' width='100%' target='_blank' style='border-radius:5px;'"
                f"src='{QONTO_CARD}'>"
                "</a>"
            ),
            "txt_intro": (
                f"Hello,<br><br>"
                f"Vous disposez de <b>{last_position}</b> chez Qonto au {date_to}."
            ),
            "title_1": emailbuilder.heading(
                "Evolution de votre trésorerie", underline=True
            ),
            "txt_1_1": emailbuilder.text(
                f"Voici un graphique de l’évolution des encaissements/décaissements depuis le {date_from}."
            ),
            "cash_graphic": emailbuilder.image(graph_img, link=graph_link),
            "txt_1_2": emailbuilder.text(
                "Cliquez sur l'image pour accéder au graphique dynamique.",
                font_size="14px",
                text_align="center",
                italic=True,
            ),
            "title_2": emailbuilder.heading(
                "Tableaux de flux par type", underline=True
            ),
            "cash_table": emailbuilder.table(
                cash_summary, **{**table_style, **table_cash}
            ),
            "title_3": emailbuilder.heading("Dernières transactions", underline=True),
            "txt_3_1": emailbuilder.text(
                f"Voici la liste des transactions effectuées ces {abs(nb_last)} derniers jours:"
            ),
            "week_table": emailbuilder.table(
                df_last, **{**table_style, **table_last_week}
            ),
            "txt_end_1": emailbuilder.text("Pour accéder au détail de vos flux:"),
            "button": emailbuilder.button(
                statement_link, "Télécharger le fichier Excel"
            ),
            "txt_end_2": emailbuilder.text(
                "Pour aller plus loin dans votre analyse découvrez tous "
                "<a href='https://github.com/jupyter-naas/awesome-notebooks/tree/master/Qonto'>les templates Qonto disponible sur Naas</a>"
                " ou contactez-nous sur hello@naas.ai pour personnaliser cet email."
            ),
            "footer_cs": emailbuilder.footer_company(naas=True),
        }
        # Generate email in html
        email_content = emailbuilder.generate(display="iframe", **content)
        return email_content



import pandas as pd
import requests
import pydash as _pd
from urllib.parse import urlencode
from datetime import datetime


class Github:
    @staticmethod
    def get_repository_url(url):
        return url.split("https://github.com/")[-1]

    def connect(self, token: str):
        # Init connect
        self.token = token

        # Init headers
        self.headers = {"Authorization": f"token {self.token}"}

        # Init end point
        self.repos = Repositories(self.headers)
        self.users = Users(self.headers)
        self.teams = Teams(self.headers)
        self.projects = Projects(self.headers)

        # Set connexion to active
        self.connected = True
        return self


class Users(Github):
    def __init__(self, headers):
        Github.__init__(self)
        self.headers = headers

    def get_profile(self, html_url, url=None):
        """
        Return a dataframe object with 20 columns:
        - LOGIN               object
        - ID                  int64
        - NODE_ID             object
        - GRAVATAR_ID         object
        - TYPE                object
        - SITE_ADMIN          bool
        - NAME                object
        - COMPANY             object
        - BLOG                object
        - LOCATION            object
        - EMAIL               object
        - HIREABLE            object
        - BIO                 object
        - TWITTER_USERNAME    object
        - PUBLIC_REPOS        int64
        - PUBLIC_GISTS        int64
        - FOLLOWERS           int64
        - FOLLOWING           int64
        - CREATED_AT          object
        - UPDATED_AT          object

        Parameters
        ----------
        html_url: str:
            User profile url from Github.
            Example : "https://github.com/SanjuEpic"
        """
        if url is None:
            user = html_url.split("github.com/")[-1].split("/")[0]
            url = f"https://api.github.com/users/{user}"

        res = requests.get(url, headers=self.headers)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            raise (e)
        res_json = res.json()

        # Dataframe
        df = pd.DataFrame([res_json])
        for col in df.columns:
            if col.endswith("url"):
                df = df.drop(col, axis=1)
            if col.endswith("_at"):
                df[col] = df[col].str.replace("T", " ").str.replace("Z", " ")
        return df


class Teams(Github):
    def __init__(self, headers):
        Github.__init__(self)
        self.headers = headers

    def get_profiles(self, url):
        """
        Return an dataframe object with 14 columns:
        - TEAM              object
        - SLUG              object
        - TEAM_DESCRIPTION  object
        - MEMBER_PROFILE    object
        - GITHUB            object
        - NAME              object
        - EMAIL             object
        - LOCATION          object
        - ORGANIZATION      object
        - BIO               object
        - LOGIN_NAME        object
        - TWITTER           object
        - CREATED_AT        object
        - UPDATED_AT        object

        Parameters
        ----------
        url: str:
            teams url from Github.
            Example : "https://github.com/orgs/jupyter-naas/teams"
        """
        # Traverses through multiple teams and all member profiles within each team
        org = url.split("https://github.com/orgs/")[-1].split("/")[0]

        member_profiles, teams, slugs, team_descriptions = [], [], [], []
        data = pd.DataFrame(
            columns=["TEAM", "SLUG", "TEAM_DESCRIPTION", "member_profile", "GITHUB"]
        )
        page = 1

        while True:
            params = {
                "state": "open",
                "per_page": "100",
                "page": page,
            }
            url = f"https://api.github.com/orgs/{org}/teams?{urlencode(params, safe='(),')}"
            res = requests.get(url, headers=self.headers)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            res_json = res.json()

            if len(res_json) == 0:
                break

            members_details = []
            for team_info in res_json:
                members_details.append(
                    (
                        team_info["name"],
                        team_info["slug"],
                        team_info["description"],
                        team_info["members_url"].strip("{/member}"),
                    )
                )

            for info in members_details:
                page_number = 1
                while True:
                    members_params = {
                        "state": "open",
                        "per_page": "100",
                        "page": page_number,
                    }

                    url = f"{info[3]}?{urlencode(members_params, safe='(),')}"
                    members = requests.get(
                        url, headers=self.headers, params=members_params
                    )

                    try:
                        members.raise_for_status()
                    except requests.HTTPError as e:
                        raise (e)
                    members_json = members.json()

                    if len(members_json) == 0:
                        break

                    for member in members_json:
                        member_profiles.append(member["url"])
                        teams.append(info[0])
                        slugs.append(info[1])
                        team_descriptions.append(info[2])

                    page_number += 1

            page += 1

        data["TEAM"], data["SLUG"], data["TEAM_DESCRIPTION"], data["member_profile"] = (
            teams,
            slugs,
            team_descriptions,
            member_profiles,
        )
        data["GITHUB"] = org

        for idx, profile in enumerate(data["member_profile"]):
            details = requests.get(profile, headers=self.headers, params=params).json()
            data.loc[idx, "NAME"], data.loc[idx, "EMAIL"], data.loc[idx, "LOCATION"] = (
                details["name"],
                details["email"],
                details["location"],
            )
            (
                data.loc[idx, "ORGANIZATION"],
                data.loc[idx, "BIO"],
                data.loc[idx, "LOGIN_NAME"],
            ) = (details["company"], details["bio"], details["login"])
            data.loc[idx, "TWITTER"], data.loc[idx, "CREATED_AT"] = (
                details["twitter_username"],
                details["created_at"],
            )
            data.loc[idx, "UPDATED_AT"] = details["updated_at"]
        return data


class Projects(Github):
    def __init__(self, headers):
        Github.__init__(self)
        self.headers = headers

    def get(self, url):
        """
        Return an dataframe object with 9 columns:
        - PROJECT_NAME            object
        - PROJECT_DESCRIPTION     object
        - PROJECT_ID              int64
        - PROJECT_CREATED_BY      object
        - PROJECT_CREATED_DATE    object
        - PROJECT_CREATED_TIME    object
        - PROJECT_UPDATED_DATE    object
        - PROJECT_UPDATED_TIME    object
        - PROJECT_COLUMNS_URL     object

        Parameters
        ----------
        url: str:
            Projects url from Github.
            Example : "https://github.com/orgs/jupyter-naas/projects"
        """
        projects_df = pd.DataFrame()
        url = "api.github.com".join(url.split("github.com"))
        page = 1
        while True:
            params = {"per_page": 100, "page": page}
            url_link = url + f"?{urlencode(params, safe='(),')}"
            res = requests.get(url_link, headers=self.headers, params=params)

            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            if len(res.json()) == 0:
                break
            res_json = res.json()

            for idx, project in enumerate(res_json):
                projects_df.loc[idx, "project_name"] = project.get("name")
                projects_df.loc[idx, "project_description"] = project.get("body")
                projects_df.loc[idx, "project_id"] = project.get("number")
                projects_df.loc[idx, "project_created_by"] = project.get("creator")[
                    "login"
                ]
                projects_df.loc[idx, "project_created_date"] = (
                    project.get("created_at").strip("Z").split("T")[0]
                )
                projects_df.loc[idx, "project_created_time"] = (
                    project.get("created_at").strip("Z").split("T")[-1]
                )
                projects_df.loc[idx, "project_updated_date"] = (
                    project.get("updated_at").strip("Z").split("T")[0]
                )
                projects_df.loc[idx, "project_updated_time"] = (
                    project.get("updated_at").strip("Z").split("T")[-1]
                )
                projects_df.loc[idx, "project_columns_url"] = project.get("columns_url")

            page += 1

        projects_df["project_id"] = projects_df.project_id.astype("int")
        return projects_df

    def get_comments_from_issues(self, url):
        """
        Returns a list of comments to a particular issue

        Parameters
        ----------
        issue comments url: str
        Example: https://github.com/jupyter-naas/awesome-notebooks/issues/359/comments
        """
        issue_comments = []

        if url.find("api.github.com") == -1:
            url = "api.github.com/repos".join(url.split("github.com"))

        comments = requests.get(url, headers=self.headers)
        try:
            comments.raise_for_status()
        except requests.HTTPError as e:
            raise (e)
        if len(comments.json()) == 0:
            return "No comments"
        else:
            for comment in comments.json():
                issue_comments.append(comment["body"])
        return issue_comments

    def get_issues(self, projects_url):
        """
        Return an dataframe object with 18 columns:
        - ISSUE_STATUS          object
        - ISSUE_STATE           object
        - LINK_TO_THE_ISSUE     object
        - ISSUE_NUMBER          int64
        - ISSUE_TITLE           object
        - ISSUE_LABELS          object
        - ISSUE_ASSIGNEES       object
        - COMMENTS_TILL_DATE    int64
        - LAST_CREATED__DATE    object
        - LAST_CREATED__TIME    object
        - LAST_UPDATED_DATE     object
        - LAST_UPDATED_TIME     object
        - STALE_ISSUE           object
        - COMMENTS              object
        - LINKED_PR_STATE       object
        - PR_ACTIVITY           object
        - PROJECT_ID            int64
        - PROJECT_NAME          object

        Parameters
        ----------
        projects_url: str:
            Projects url from Github.
            Example : "https://github.com/orgs/jupyter-naas/projects"
        """
        df_projects = self.get(projects_url)
        df_issues = pd.DataFrame(columns=["issue_status", "issue_state"])

        # Gets info from columns present in our roadmap for all active projects
        for _, project in df_projects.iterrows():
            columns = requests.get(
                project["project_columns_url"], headers=self.headers
            ).json()
            issue_status, issue_urls = [], []
            for column in columns:
                page = 1
                while True:
                    params = {"per_page": 100, "page": page}
                    card_url = column["cards_url"] + f"?{urlencode(params, safe='(),')}"
                    issues = requests.get(card_url, headers=self.headers)
                    try:
                        issues.raise_for_status()
                    except requests.HTTPError as e:
                        raise (e)
                    if len(issues.json()) == 0:
                        break
                    for issue in issues.json():
                        if issue.get("content_url") is not None:
                            issue_urls.append(issue.get("content_url"))
                            issue_status.append(column["name"])
                    page += 1

            df_issues["issue_status"] = issue_status

            for idx, url in enumerate(issue_urls):
                issue = requests.get(url, headers=self.headers)
                try:
                    issue.raise_for_status()
                except requests.HTTPError as e:
                    raise (e)
                issue = issue.json()
                # information to be extracted are below
                (
                    df_issues.loc[idx, "link_to_the_issue"],
                    df_issues.loc[idx, "issue_number"],
                ) = (issue["html_url"], issue["number"])
                df_issues.loc[idx, "issue_title"], df_issues.loc[idx, "issue_state"] = (
                    issue["title"],
                    issue["state"],
                )

                labels = []
                for label in issue["labels"]:
                    labels.append(label.get("name"))
                df_issues.loc[idx, "issue_labels"] = ", ".join(labels)

                assigned = []
                for assignee in issue["assignees"]:
                    assigned.append(assignee.get("login"))
                if assigned == []:
                    df_issues.loc[idx, "issue_assignees"] = "None"
                else:
                    df_issues.loc[idx, "issue_assignees"] = ", ".join(assigned)

                df_issues.loc[idx, "comments_till_date"] = issue["comments"]

                df_issues.loc[idx, "last_created_date"] = (
                    issue.get("created_at").strip("Z").split("T")[0]
                )
                df_issues.loc[idx, "last_created_time"] = (
                    issue.get("created_at").strip("Z").split("T")[-1]
                )
                df_issues.loc[idx, "last_updated_date"] = (
                    issue.get("updated_at").strip("Z").split("T")[0]
                )
                df_issues.loc[idx, "last_updated_time"] = (
                    issue.get("updated_at").strip("Z").split("T")[-1]
                )

                if df_issues.loc[idx, "issue_status"] != "Backlog":
                    date_format = "%Y-%m-%d"
                    delta = datetime.now() - datetime.strptime(
                        df_issues.loc[idx, "last_updated_date"], date_format
                    )
                    df_issues.loc[
                        idx, "stale_issue"
                    ] = f"No activity since {delta.days} days"
                else:
                    df_issues.loc[idx, "stale_issue"] = "None"

                df_issues.loc[idx, "comments"] = str(
                    self.get_comments_from_issues(issue["comments_url"])
                )

                try:
                    pr = requests.get(
                        issue.get("pull_request")["url"], headers=self.headers
                    ).json()
                    df_issues.loc[idx, "linked_pr_state"] = pr.get("state")

                    date_format = "%Y-%m-%d"
                    delta = datetime.now() - datetime.strptime(
                        pr.get("updated_at").split("T")[0], date_format
                    )
                    df_issues.loc[
                        idx, "PR_activity"
                    ] = f"No activity since {delta.days} days"

                except Exception:
                    df_issues.loc[idx, "linked_pr_state"] = "None"
                    df_issues.loc[idx, "PR_activity"] = "None"

            df_issues["project_id"] = [
                df_projects.project_id.values[0]
            ] * df_issues.shape[0]
            df_issues["project_name"] = [
                df_projects.project_name.values[0]
            ] * df_issues.shape[0]
            df_issues["issue_number"] = df_issues["issue_number"].apply(
                lambda x: int(x)
            )
            df_issues["comments_till_date"] = df_issues["comments_till_date"].apply(
                lambda x: int(x)
            )

        return df_issues


class Repositories(Github):
    def __init__(self, headers):
        Github.__init__(self)
        self.headers = headers

    def get_commits(
        self, url, author=None, since=None, until=None, per_page=100, page=1
    ):
        """
        url: string
        Repository url.

        author: string
        GitHub login or email address by which to filter by commit author.

        since: string
        Only show notifications updated after the given time. This is a timestamp in ISO 8601 format: YYYY-MM-DDTHH:MM:SSZ.

        until: string
        Only commits before this date will be returned. This is a timestamp in ISO 8601 format: YYYY-MM-DDTHH:MM:SSZ.

        per_page: integer
        The number of results per page (max 100).
        Default: 30

        page: integer
        Page number of the results to fetch.
        Default: 1

        Return an dataframe object with 11 columns:
        - ID                   object
        - MESSAGE              object
        - AUTHOR_DATE          date
        - AUTHOR_NAME          object
        - AUTHOR_EMAIL         object
        - COMMITTER_DATE       date
        - COMMITTER_NAME       object
        - COMMITTER_EMAIL      object
        - COMMENTS_COUNT       int
        - VERIFICATION_REASON  object
        - VERIFICATION_STATUS  object

        Parameters
        ----------
        repository: str:
            Repository url from Github.
            Example : "https://github.com/jupyter-naas/awesome-notebooks"
        """
        # Get organisation and repository from url
        repository = Github.get_repository_url(url)

        # Get commits
        commits = []
        while True:
            params = {
                "state": "open",
                "per_page": per_page,
                "page": page,
            }
            if author:
                params["author"] = author
            if since:
                params["since"] = since
            if until:
                params["until"] = until
            url = f"https://api.github.com/repos/{repository}/commits?{urlencode(params, safe='(),')}"
            res = requests.get(url, headers=self.headers)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            res_json = res.json()

            if len(res_json) == 0:
                break
            for r in res_json:
                commit = {
                    "ID": _pd.get(r, "sha"),
                    "MESSAGE": _pd.get(r, "commit.message"),
                    "AUTHOR_DATE": _pd.get(r, "commit.author.date")
                    .replace("T", " ")
                    .replace("Z", ""),
                    "AUTHOR_NAME": _pd.get(r, "commit.author.name"),
                    "AUTHOR_EMAIL": _pd.get(r, "commit.author.email"),
                    "COMMITTER_DATE": _pd.get(r, "commit.committer.date")
                    .replace("T", " ")
                    .replace("Z", ""),
                    "COMMITTER_NAME": _pd.get(r, "commit.committer.name"),
                    "COMMITTER_EMAIL": _pd.get(r, "commit.committer.email"),
                    "COMMENTS_COUNT": _pd.get(r, "commit.comment_count"),
                    "VERIFICATION_REASON": _pd.get(r, "commit.verification.reason"),
                    "VERIFICATION_STATUS": _pd.get(r, "commit.verification.verified"),
                }
                commits.append(commit)
            page += 1
        # Return dataframe
        df = pd.DataFrame(commits)
        df["AUTHOR_DATE"] = pd.to_datetime(df["AUTHOR_DATE"])
        df["COMMITTER_DATE"] = pd.to_datetime(df["COMMITTER_DATE"])
        return df

    def get_stargazers(self, url):
        """
        Return an dataframe object with 6 columns:
        - LOGIN       object
        - ID          int64
        - URL         object
        - TYPE        object
        - SITE_ADMIN  bool
        - STARRED_AT  object

        Parameters
        ----------
        repository: str:
            Repository url from Github.
            Example : "https://github.com/jupyter-naas/awesome-notebooks"
        """
        # Get organisation and repository from url
        repository = Github.get_repository_url(url)

        # Custom headers
        headers = self.headers
        headers["Accept"] = "application/vnd.github.v3.star+json"

        df = pd.DataFrame()
        page = 1
        while True:
            params = {
                "per_page": "100",
                "page": page,
            }
            url = f"https://api.github.com/repos/{repository}/stargazers?{urlencode(params, safe='(),')}"
            res = requests.get(url, headers=headers)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            res_json = res.json()

            if len(res_json) == 0:
                break
            for json in res_json:
                starred_at = json.get("starred_at")
                user = json.get("user")
                tmp = pd.DataFrame([user])
                tmp["starred_at"] = starred_at
                df = pd.concat([df, tmp], axis=0)
            page += 1

        # Cleaning
        for col in df.columns:
            if col.endswith("_url") or col.endswith("_id"):
                df = df.drop(col, axis=1)
            if col.endswith("_at"):
                df[col] = df[col].str.replace("T", " ").str.replace("Z", " ")
        df.columns = df.columns.str.upper()
        return df

    def get_comments_from_issues(self, url):
        """
        Returns a list of comments to a particular issue

        Parameters
        ----------
        issue comments url: str
        Example: https://github.com/jupyter-naas/awesome-notebooks/issues/359/comments
        """
        issue_comments = []

        if url.find("api.github.com") == -1:
            url = "api.github.com".join(url.split("github.com"))

        comments = requests.get(url, headers=self.headers)
        try:
            comments.raise_for_status()
        except requests.HTTPError as e:
            raise (e)
        if len(comments.json()) == 0:
            return "No comments"
        else:
            for comment in comments.json():
                issue_comments.append(comment["body"])
        return issue_comments

    def get_issues(self, url):
        """
        Return an dataframe object with 15 columns:
        - LINK_TO_THE_ISSUE      object
        - ISSUE_NUMBER           int64
        - ISSUE_TITLE            object
        - ISSUE_STATE            object
        - ISSUE_ID               int64
        - ISSUE_LABELS           object
        - ISSUE_ASSIGNEES        object
        - COMMENTS_TILL_DATE     int64
        - LAST_CREATED_DATE      object
        - LAST_CREATED_TIME      object
        - LAST_UPDATED_DATE      object
        - LAST_UPDATED_TIME      object
        - COMMENTS               object
        - LINKED_PR_STATE        object
        - PR_ACTIVITY            object

        Parameters
        ----------
        repository: str:
            Repository url from Github.
            Example : "https://github.com/jupyter-naas/awesome-notebooks"
        """
        # Get organisation and repository from url
        repository = Github.get_repository_url(url)

        df = pd.DataFrame()
        page, idx = 1, 0
        while True:
            params = {
                "per_page": "100",
                "page": page,
            }
            url = f"https://api.github.com/repos/{repository}/issues?{urlencode(params, safe='(),')}"
            res = requests.get(url, headers=self.headers)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            res_json = res.json()
            if len(res_json) == 0:
                break

            for issue in res_json:
                df.loc[idx, "link_to_the_issue"], df.loc[idx, "issue_number"] = (
                    issue["html_url"],
                    issue["number"],
                )
                df.loc[idx, "issue_title"], df.loc[idx, "issue_state"] = (
                    issue["title"],
                    issue["state"],
                )
                df.loc[idx, "issue_id"] = issue["id"]
                labels = []
                for label in issue["labels"]:
                    labels.append(label.get("name"))
                if labels == []:
                    df.loc[idx, "issue_labels"] = "None"
                else:
                    df.loc[idx, "issue_labels"] = ", ".join(labels)

                assigned = []
                for assignee in issue["assignees"]:
                    assigned.append(assignee.get("login"))
                if assigned == []:
                    df.loc[idx, "issue_assignees"] = "None"
                else:
                    df.loc[idx, "issue_assignees"] = ", ".join(assigned)

                df.loc[idx, "comments_till_date"] = issue["comments"]

                df.loc[idx, "last_created_date"] = (
                    issue.get("created_at").strip("Z").split("T")[0]
                )
                df.loc[idx, "last_created_time"] = (
                    issue.get("created_at").strip("Z").split("T")[-1]
                )
                df.loc[idx, "last_updated_date"] = (
                    issue.get("updated_at").strip("Z").split("T")[0]
                )
                df.loc[idx, "last_updated_time"] = (
                    issue.get("updated_at").strip("Z").split("T")[-1]
                )

                df.loc[idx, "comments"] = str(
                    self.get_comments_from_issues(issue["comments_url"])
                )

                try:
                    pr = requests.get(
                        issue.get("pull_request")["url"], headers=self.headers
                    ).json()
                    df.loc[idx, "linked_pr_state"] = pr.get("state")

                    date_format = "%Y-%m-%d"
                    delta = datetime.now() - datetime.strptime(
                        df.loc[idx, "last_updated_date"], date_format
                    )
                    df.loc[idx, "PR_activity"] = f"No activity since {delta.days} days"

                except Exception:
                    df.loc[idx, "linked_pr_state"] = "None"
                    df.loc[idx, "PR_activity"] = "None"

                idx += 1

            page += 1

        df["issue_id"] = df.issue_id.astype("int")
        df["comments_till_date"] = df.comments_till_date.astype("int")
        df["issue_number"] = df.issue_number.astype("int")

        return df

    def get_pulls(self, url):
        """
        Return an dataframe object with 15 columns:
        - ID                      int64
        - ISSUE_URL               object
        - PR_NUMBER               int64
        - PR_STATE                object
        - TITLE                   object
        - FIRST_CREATED_DATE      object
        - FIRST_CREATED_TIME      object
        - LAST_UPDATED_DATE       object
        - LAST_UPDATED_TIME       object
        - COMMITS_URL             object
        - REVIEW_COMMENTS_URL     object
        - ISSUE_COMMENTS_URL      object
        - ASSIGNEES               object
        - REQUESTED_REVIEWERS     object
        - PR_ACTIVITY             object

        Parameters
        ----------
        repository: str:
            Repository url from Github.
            Example : "https://github.com/jupyter-naas/awesome-notebooks"
        """
        # Get organisation and repository from url
        repository = Github.get_repository_url(url)

        df = pd.DataFrame()
        page = 1
        while True:
            params = {
                "per_page": "100",
                "page": page,
            }
            url = f"https://api.github.com/repos/{repository}/pulls?{urlencode(params, safe='(),')}"
            res = requests.get(url, headers=self.headers)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                raise (e)
            res_json = res.json()
            if len(res_json) == 0:
                break

            for idx, r in enumerate(res_json):
                if r.get("state") == "open":
                    df.loc[idx, "id"] = r.get("id")
                    df.loc[idx, "issue_url"] = r.get("issue_url")
                    df.loc[idx, "PR_number"] = r.get("number")
                    df.loc[idx, "PR_state"] = "open"
                    df.loc[idx, "Title"] = r.get("title")

                    df.loc[idx, "first_created_date"] = (
                        r.get("created_at").strip("Z").split("T")[0]
                    )
                    df.loc[idx, "first_created_time"] = (
                        r.get("created_at").strip("Z").split("T")[-1]
                    )
                    df.loc[idx, "last_updated_date"] = (
                        r.get("updated_at").strip("Z").split("T")[0]
                    )
                    df.loc[idx, "last_updated_time"] = (
                        r.get("updated_at").strip("Z").split("T")[-1]
                    )

                    df.loc[idx, "commits_url"] = r.get("commits_url")
                    df.loc[idx, "review_comments_url"] = r.get("review_comments_url")
                    df.loc[idx, "issue_comments_url"] = r.get("comments_url")

                    assignees_lst, reviewers_lst = [], []
                    for assignee in r.get("assignees"):
                        assignees_lst.append(assignee.get("login"))
                    for reviewer in r.get("requested_reviewers"):
                        reviewers_lst.append(reviewer.get("login"))

                    if assignees_lst == []:
                        df.loc[idx, "assignees"] = "None"
                    elif assignees_lst:
                        df.loc[idx, "assignees"] = ", ".join(assignees_lst)

                    if reviewers_lst == []:
                        df.loc[idx, "requested_reviewers"] = "None"
                    elif reviewers_lst:
                        df.loc[idx, "requested_reviewers"] = ", ".join(reviewers_lst)

                    date_format = "%Y-%m-%d"
                    delta = datetime.now() - datetime.strptime(
                        df.loc[idx, "last_updated_date"], date_format
                    )
                    df.loc[idx, "PR_activity"] = f"No activity since {delta.days} days"

                df["PR_number"] = df.PR_number.astype("int")
                df.id = df.id.astype("int")

            page += 1

        return df



import requests
import pandas as pd
import datetime
from youtube_transcript_api import YouTubeTranscriptApi
from transformers import pipeline
from tqdm import tqdm
import pydash as _


YOUTUBE_API_URL = "https://www.googleapis.com/youtube/v3"


class Youtube:
    def __init__(self):
        self.transcript = Transcript()

    def connect(self, api_key: str):
        # Init lk attribute
        self.api_key = api_key

        # Init cookies
        self.base_params = {"key": self.api_key}

        # Init end point
        self.channel = Channel(self.base_params)
        self.video = Video(self.base_params)

        # Set connexion to active
        self.connected = True
        return self

    def get_video_id_from_url(self, video_url):
        video_id = video_url.split("watch?v=")[-1].split("&")[0]
        return video_id


class Channel(Youtube):
    def __init__(self, base_params):
        Youtube.__init__(self)
        self.base_params = base_params

    def __get_channel_id_from_url(self, channel_url):
        channel_id = channel_url.split("channel/")[-1].split("/")[0]
        return channel_id

    def __get_uploads_id_from_channel(self, channel_url):
        channel_id = self.__get_channel_id_from_url(channel_url)
        params = {"part": "contentDetails", "id": channel_id}
        params.update(self.base_params)
        res = requests.get(f"{YOUTUBE_API_URL}/channels", params=params)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        playlist_id = _.get(
            res.json(), "items[0].contentDetails.relatedPlaylists.uploads"
        )
        return playlist_id

    def get_statistics(self, channel_url):
        if "www.youtube.com/channel" not in channel_url:
            return (
                "❌ Channel url not valid. "
                "Please get url with channel id. "
                "It must start with 'www.youtube.com/channel'"
            )
        channel_id = self.__get_channel_id_from_url(channel_url)
        params = {"part": "statistics,snippet", "id": channel_id}
        params.update(self.base_params)
        res = requests.get(f"{YOUTUBE_API_URL}/channels", params=params)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        stats = _.get(res.json(), "items[0]")
        data = {
            "ID": _.get(stats, "id"),
            "URL": channel_url,
            "TITLE": _.get(stats, "snippet.title"),
            "COUNTRY": _.get(stats, "snippet.country"),
            "DESCRIPTION": _.get(stats, "snippet.description"),
            "VIEWCOUNT": _.get(stats, "statistics.viewCount"),
            "SUBSCRIBERCOUNT": _.get(stats, "statistics.subscriberCount"),
            "VIDEOCOUNT": _.get(stats, "statistics.videoCount"),
            "THUMBNAILS": _.get(stats, "snippet.thumbnails.high.url"),
            "PUBLISHEDAT": _.get(stats, "snippet.publishedAt"),
        }
        df = pd.DataFrame([data])
        return df

    def get_uploads(self, channel_url, number=100):
        if "www.youtube.com/channel" not in channel_url:
            return (
                "❌ Channel url not valid. "
                "Please get url with channel id. "
                "It must start with 'www.youtube.com/channel'"
            )
        playlist_id = self.__get_uploads_id_from_channel(channel_url)
        videos = []
        data = []
        nextPageToken = None
        while len(videos) < number:
            maxResults = 50 if (number - len(videos)) > 50 else (number - len(videos))
            params = {
                "part": "snippet,contentDetails",
                "maxResults": maxResults,
                "playlistId": playlist_id,
            }
            if nextPageToken:
                params["PageToken"] = nextPageToken
            params.update(self.base_params)
            res = requests.get(f"{YOUTUBE_API_URL}/playlistItems", params=params)
            try:
                res.raise_for_status()
            except requests.HTTPError as e:
                return e
            videos += _.get(res.json(), "items")
            nextPageToken = _.get(res.json(), "nextPageToken")

            for video in videos:
                meta = {
                    "CHANNEL_ID": _.get(video, "snippet.channelId"),
                    "PLAYLIST_ID": _.get(video, "snippet.playlistId"),
                    "VIDEO_ID": _.get(video, "snippet.resourceId.videoId"),
                    "VIDEO_TITLE": _.get(video, "snippet.title", ""),
                    "VIDEO_DESCRIPTION": _.get(video, "snippet.description", ""),
                    "VIDEO_PUBLISHEDAT": _.get(
                        video, "contentDetails.videoPublishedAt"
                    ),
                    "VIDEO_THUMBNAILS": _.get(
                        video, "snippet.thumbnails.maxres.url", ""
                    ),
                }
                data.append(meta)
            if nextPageToken is None:
                break
        df = pd.DataFrame(data)
        df["PUBLISHEDAT"] = pd.to_datetime(df["VIDEO_PUBLISHEDAT"])
        return df


class Video(Youtube):
    def __init__(self, base_params):
        Youtube.__init__(self)
        self.base_params = base_params

    def get_statistics(self, video_url):
        video_id = Youtube.get_video_id_from_url(self, video_url)
        params = {"part": "statistics,snippet,contentDetails", "id": video_id}
        params.update(self.base_params)
        res = requests.get(f"{YOUTUBE_API_URL}/videos", params=params)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            return e
        stats = _.get(res.json(), "items[0]")

        # Calcul duration
        duration = _.get(stats, "contentDetails.duration")

        def duration_to_seconds(duration):
            total_seconds = 0
            duration = duration.replace("P", "").replace("T", "")

            def calcul_duration(string, variable):
                result = 0
                if variable in duration:
                    result = string.split(variable)[0]
                    try:
                        result = int(result[-2:])
                    except ValueError:
                        result = int(result[-1:])
                return result

            # Get duration split
            seconds = calcul_duration(duration, "S")
            minutes = calcul_duration(duration, "M")
            hours = calcul_duration(duration, "H")
            days = calcul_duration(duration, "D")

            total_seconds = (
                seconds + (minutes * 60) + (hours * 3600) + (days * 3600 * 24)
            )
            return total_seconds

        duration_seconds = duration_to_seconds(duration)
        duration_final = str(datetime.timedelta(seconds=duration_seconds))

        data = {
            "CHANNEL_ID": _.get(stats, "snippet.channelId"),
            "CHANNEL_TITLE": _.get(stats, "snippet.channelTitle"),
            "ID": _.get(stats, "id"),
            "URL": video_url,
            "TITLE": _.get(stats, "snippet.title"),
            "DESCRIPTION": _.get(stats, "snippet.description"),
            "VIEWCOUNT": _.get(stats, "statistics.viewCount"),
            "LIKECOUNT": _.get(stats, "statistics.likeCount"),
            "DISLIKECOUNT": _.get(stats, "statistics.dislikeCount"),
            "FAVORITECOUNT": _.get(stats, "statistics.favoriteCount"),
            "COMMENTCOUNT": _.get(stats, "statistics.commentCount"),
            "DURATION": duration_final,
            "DURATION_SECONDS": duration_seconds,
            "THUMBNAILS": _.get(stats, "snippet.thumbnails.high.url"),
            "PUBLISHEDAT": _.get(stats, "snippet.publishedAt"),
        }
        df = pd.DataFrame([data])
        df["PUBLISHEDAT"] = pd.to_datetime(df["PUBLISHEDAT"])
        return df


class Transcript:
    def __text_to_chunks(self, text, chunk_size):
        chunks = []
        count = 0
        text_len = len(text)

        while count * chunk_size < text_len:
            chunks.append(
                text[count * chunk_size : count * chunk_size + chunk_size]  # noqa: E203
            )
            count += 1
        return chunks

    def __pipeline_summarization(self, text):
        chunks = self.__text_to_chunks(text, 1024)
        summaries = []
        summarization = pipeline("summarization")

        for chunk in tqdm(chunks):
            summaries.append(summarization(chunk)[0]["summary_text"])
        return summaries

    def get(self, video_url):
        video_id = Youtube.get_video_id_from_url(self, video_url)
        try:
            json = YouTubeTranscriptApi.get_transcript(video_id)
        except Exception as e:
            return e
        return json

    def summarize(self, video_url):
        json = self.get(video_url)
        para = ""
        for i in json:
            para += i["text"]
            para += " "
        print("🎬 Transcript:", len(para), "characters")
        summaries = self.__pipeline_summarization(para)
        summary = " ".join(summaries)
        summary = (
            summary.replace(" .", ".")
            .replace("\xa0", " ")
            .replace("\n", " ")
            .replace("  ", " ")
        )
        print("👉 Summary:", len(summary), "characters")
        return summary.strip()



from naas_drivers.driver import InDriver, OutDriver
import pandas as pd  # noqa: F401
import sys
import time
from pymongo import MongoClient


filter_system = {"name": {"$regex": r"^(?!system\.)"}}


class Mongo(InDriver, OutDriver):
    """Mongo lib"""

    __client = None
    # CONNECT TO MONGODB
    # - Function name : mongo_connect
    # - Arguments : host, port, username, password
    # - Value return : MongoClient
    # - Message :
    # --- Success => "Successfully connected to MongoDB"
    # --- Failed => "Error connecting to MongoDB. Please check configuration"

    def get_client(self):
        return self.__client

    def connect(
        self, mongo_host, mongo_port=None, mongo_username=None, mongo_password=None
    ):
        if mongo_port and mongo_username and mongo_password:
            self.__client = MongoClient(
                mongo_host,
                mongo_port,
                username=mongo_username,
                password=mongo_password,
            )
        else:
            self.__client = MongoClient(mongo_host)

        self.__client.server_info()
        print("Successfully connected to MongoDB")
        self.connected = True
        return self

    # SAVE DF IN MONGODB
    # - Function name : save_df
    # - Arguments : df, collection_name, db_name, cancel & replace => default = False)
    # - Value return : None
    # - Message :
    # --- Success => "Dataframe successfully save in MongoDB"
    # --- Failed => "Failed to save in MongoDB. Please ask Bob for help"

    def send(self, df, collection_name, db_name, replace=False):
        self.check_connect()
        start_time = time.time()
        # Init collection
        mongo_db = self.__client[db_name]
        df_collection = mongo_db[collection_name]

        try:
            # Delete collection if already exist
            if replace:
                for collection in mongo_db.list_collection_names(filter=filter_system):
                    if collection == collection_name:
                        df_collection.drop()

            # Check size => monogdb will only allow 16MB data at a time to be inserted
            chunk_size = round(sys.getsizeof(df) / 16793600)
            if chunk_size > 0:
                my_list = df.to_dict("records")
                lenght = len(my_list)
                ran = list(range(lenght))
                steps = ran[chunk_size::chunk_size]
                steps.extend([lenght])

                # Inser chunks of the dataframe
                i = 0
                for j in steps:
                    df_collection.insert_many(my_list[i:j])
                    i = j
            else:
                df_collection.insert_many(df.to_dict(orient="records"))
            print(
                f"Dataframe {collection_name} successfully save in database {db_name} in MongoDB. Time: --- %s secnds ---"
                % (time.time() - start_time)
            )
        except Exception as e:
            print("Failed to save in MongoDB.")
            print(e.__doc__)
            print(str(e))

    # Get DF IN MONGODB
    # - Function name : df_to_mongo
    # - Arguments : collection_name, db_name, filters => default = {}
    # - Value return : df
    # - Message :
    # --- Success => "Dataframe successfully save in MongoDB"
    # --- Failed => "Failed to save in MongoDB. Please ask Bob for help"

    def get(self, collection_name, db_name, filters={}):
        self.check_connect()
        # Init collection
        mongo_db = self.__client[db_name]
        df_collection = mongo_db[collection_name]

        try:
            # Read
            df = pd.DataFrame(list(df_collection.find(filters)))
            return df
        except Exception as e:
            print("Failed to read MongoDB")
            print(e.__doc__)
            print(str(e))
            return None
